# array.tcl --
#
#	Extended array handling code. The two_array command knows the
#	options "load" and "save" and uses a more human-readable
#	formatting (key:value pairs one on a row, with several characters
#	being masked)
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/array.tcl,v 1.15 1999/10/26 13:12:13 fjl Exp $

package provide twoArray 1.0
package require twoUtil

namespace eval ::two::array {
    namespace export two_array
    namespace import ::two::util::two_now
}

# two_array --
#
#	wrapper around tcl array, which is exteded with a
#	load and save command.
#
# Arguments:
#	see manpage

proc ::two::array::two_array {opt args} {
    switch -exact -- $opt {
        load {
	    return [eval LoadArray $args]
	}
	save {
	    return [eval SaveArray $args]
	}
	default {
	    return [uplevel array $opt $args]
	}
    }
    return
}


# ::two::array::LoadArray --
#
#	Read and parse simple array file: set elements of array "arrayName"
#	with "key:value" pairs from array file "fileName".
#	If force is set and no array file is found then there will be an error
#	message on stderr, an abort message on stdout and the program will exit.
#
# Arguments:
#	fileName	Name of array file
#	arrayName	name of array which will contain the configuration
#	force		1 if config file is required to be already there
#
# Results:
#	1 if successfull, error otherwise

proc ::two::array::LoadArray {fileName arrayName {force 0}} {
    upvar 2 $arrayName a

    if [ catch {
        set arrayFile [two_readFile -nonewline $fileName]
    } errorMsg ] {
        if {$force} {
            # We were told that the array file must exist
            puts stderr "Failed to read `$fileName' at [two_now]\n  -> `$errorMsg'"
            puts "aborted: [two_now]"
            exit
        } else {
            return 0
        }
    } else {
        # Read config file
	foreach line [split $arrayFile \n] {
	    foreach {key val} [split $line ":"] {break}
	    # undo coding
	    regsub -all "%n" $key "\n" key
	    regsub -all "%n" $val "\n" val
	    regsub -all "%d" $key ":"  key
	    regsub -all "%d" $val ":"  val
	    regsub -all "%p" $key "%"  key
	    regsub -all "%p" $val "%"  val
	    # set array
            set a($key) $val
	}
    }

    return 1
}

# ::two::array::SaveArray --
#
#	Save array to file: the elements of array "arrayName"
#	will be written as "key:value" pairs to array file "fileName".
#	Will put an error message on stderr if writing failes.
#
# Arguments:
#	fileName	Name of array file
#	configName	name of array which will contain the configuration
#
# Results:
#	1 if successfull, error otherwise

proc ::two::array::SaveArray {fileName arrayName} {
    upvar 2 $arrayName a
    set ok 1

    if [ catch {
        set arrayFile [open $fileName w]
        foreach key [lsort [array names a]] {
	    set val $a($key)
	    # coding: "%" => "%p" and ":" => "%d", "\n" => "%n"
	    # so only one ":" per line
	    regsub -all "%" $key  "%p" key
	    regsub -all "%" $val  "%p" val
	    regsub -all ":" $key  "%d" key
	    regsub -all ":" $val  "%d" val
	    regsub -all "\n" $key "%n" key
	    regsub -all "\n" $val "%n" val
	    # write to file
            puts $arrayFile "$key:$val"
        }
        close $arrayFile
    } errorMsg ] {
        # error message if write failed
        set ok 0
    	puts stderr "Failed to write `$fileName' at [two_now]\n  -> `$errorMsg'"
    }
    return $ok
}

namespace import ::two::array::*
