# autosb.tcl --
#
#	Automatic scrollbar. Behaves like a regular scrollbar, but
#	disappears if not needed.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/autosb.tcl,v 1.19 1999/10/26 13:12:13 fjl Exp $

package provide twoAutosb 1.0
package require twoUtil

option add *TwoAutoSb.borderWidth 0 widgetDefault
option add *TwoAutoSb.takeFocus   0 widgetDefault

namespace eval ::two::autosb {
    namespace export two_autosb
    variable Autosb
    
    # variable Autosb
    #
    # $w,shown -	true if scrollbar is on
    # $w,blockS -	true if scrollbar has recently been added and should
    #		stay the same for some time to avoid flicker
    # $w,blockH -	dito for recently hidden scrollbar
}

# ::two::autosb::two_autosb --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.sb"); the original widget procedure
#	is renamed to "_two$w" (like "_cb.tl.sb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::autosb::two_autosb {w args} {
    variable Autosb

    # Initialise the globals
    array set Autosb [list \
        $w,shown  0 \
        $w,blockS 0 \
        $w,blockH 0 \
    ]

    # Container frame
    frame $w -class TwoAutoSb

    # the original scrollbar
    uplevel [list scrollbar $w.sb] $args
    
    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Autosb
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::autosb::Widget --
#
#	replacement for widget command.
#	One tricky part here is that we block changes in the state
#	of the autosb (i.e. show/hide) for 200 ms after each change.
#	This is needed e.g. for horizintal scrollbars in text widgets:
#	Take a horizontal scrollbar whose heigth is at least that of
#	a text line, so showing the scrollbar will cover the bottom
#	visible line. Make this bottom line the only one which is
#	long enough to require a horizontal scrollbar. So, showing
#	the scrollbar removes the need for a scrollbar, and making it
#	disappear again creates the need for a scrollbar. Bummer -
#	infinit flicker. The 200 ms blocking time is just a wild
#	guess, on a very slow machine (or slow X connection, maybe
#	with -sync) this might not work. But in those cases the whole
#	application would be ratther "jumpy".
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::autosb::Widget {w opt args} {
    variable Autosb

    # See if this option is for us...
    if {[string match "set" $opt] && [llength $args] == 2} {
	set off [expr {[lindex $args 0] == 0 && [lindex $args 1] == 1}]
	if {$off && $Autosb($w,shown) && !$Autosb($w,blockS)} {
	    pack forget $w.sb
	    _two$w configure -width 1 -height 1
	    set Autosb($w,shown)  0
	    set Autosb($w,blockS) 1
	    after 200 [list set [namespace current]::Autosb($w,blockS) 0]
	} elseif {!$off && !$Autosb($w,shown) && !$Autosb($w,blockH)} {
	    pack $w.sb -fill both -expand 1
	    set Autosb($w,shown)  1
	    set Autosb($w,blockH) 1
	    after 200 [list set [namespace current]::Autosb($w,blockH) 0]
	}
    }
    return [uplevel [list $w.sb $opt] $args]
}

# ::two::autosb::_Destroy --
#
#	Call this procedure when the autosb widget is destroyed.
#	Will destroy the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::autosb::_Destroy {w} {
    variable Autosb

    # clear Autosb
    foreach key [array names Autosb "$w,*"] {
        unset Autosb($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

namespace import ::two::autosb::*
