# editor.tcl --
#
#	Editor widget. Consists of 2 two_text widgets, managed by a
#	two_splitter to make it "splittable"
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/editor.tcl,v 1.63 1999/11/08 20:26:18 fjl Exp $

package provide twoEditor 1.0
package require twoUtil
package require twoSplitter
package require twoText
package require twoAutosb

# Create fonts (catch as they might already exist)
catch {font create twoEditorS -family courier -size 10}
catch {font create twoEditorC -family courier -size 10 -slant italic}
catch {font create twoEditorK -family courier -size 10 -weight bold}

namespace eval ::two::editor {
    namespace export two_editor
    variable Editor
    
    namespace import ::two::util::two_parseArgs
    
    # variable Editor
    #
    # nonce -	nonce to give any new file a new number
    # matchId -	id of after event handler for removing match highlighting
    # hl,active -	true if HighlightLoop is active
    # hl,widgets -	List of text widgets which currently are being highlighted
    # hl,startLen -	Length of match for highlight start delimiter
    # hl,stopLen -	Length of match for highlight stop delimiter
    # $w,file -	name of current file
    # $w,mtime -	file mtime of file on disk
    # $w,split -	true if editor is split
    # $w,new -	true if editor holds a new (i.e. unnamed) file
    # $w,children -	path names of all children of this editor (frames...)
    # $w,scrollx -	true if a horizontal scrollbar is used
    # $w,closecmd -	command that is to be called on close
    # $w,saveerrcmd -	command that is to be called if saving fails
    # $w,openerrcmd -	command that is to be called if open fails
    # $w,active -	active part (has focus or had it recently)
    # $w,style -    look of the splitter used by the editor {windows|unix}
    # $w,match -	show matching parantheses if value is true
    # $w,high -	do keyword highlighting
    # $w,comment -	list of pairs of comment delimiters
    # $w,string -	list of string delimiters
    # $w,keywords -	list of keywords to highlight
    # $w,keywordRE -	regular expression describing "keywords"
    # $w,startRE -	Regular expression to find start of comments and strings
    # $w,stop,$s -	End delimiter of comment or string according to start $s
    # $w,tag,$s -	What tag to add according to start delimiter $s
    # $w,id -	Increased whenever highlighting is started
    
    array set Editor "
        nonce	0
        bg		{}
        hl,active	0
        hl,widgets	{}
        style		$tcl_platform(platform)
        matchId		0
    "
}

# ::two::editor::two_editor --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.ed"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.ed") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	editor will create two two_text widgets and a pane.
#	Arguments are passed to both two_text widgets.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::editor::two_editor {w args} {
    variable Editor

    # Initialise the globals
    array set Editor "
        $w,file		{}
	$w,buffer	{}
	$w,mtime	0
        $w,split	0
        $w,new		1
	$w,scrollx	1
        $w,closecmd	{}
        $w,saveerrcmd	{}
        $w,openerrcmd	{}
	$w,active	$w.1.txt
	$w,children	$w.1
	$w,style        $Editor(style)
	$w,match	1
	$w,highlight	0
	$w,comment	{}
	$w,string	{}
	$w,keywords	{}
	$w,keywordRE	{}
    "

    # See if this option is for us...
    two_parseArgs args Editor($w,split)      "-split"
    two_parseArgs args Editor($w,scrollx)    "-scrollx"
    two_parseArgs args Editor($w,closecmd)   "-closecmd"
    two_parseArgs args Editor($w,saveerrcmd) "-saveerrcmd"
    two_parseArgs args Editor($w,openerrcmd) "-openerrcmd"
    two_parseArgs args Editor($w,file)       "-file"
    two_parseArgs args Editor($w,buffer)     "-buffer"
    two_parseArgs args Editor($w,style)      "-style"
    two_parseArgs args Editor($w,match)      "-match"
    two_parseArgs args Editor($w,highlight)  "-hightlight"
    two_parseArgs args Editor($w,comment)    "-comment"
    two_parseArgs args Editor($w,string)     "-string"
    two_parseArgs args Editor($w,keywords)   "-keywords"
    two_parseArgs args Editor($w,keywordRE)  "-keywordRE"

    # Container frame
    # Explicitly takes the focus to be able to
    # directly go to the proper child
    frame $w -class TwoEditor -bd 0 -width 500 -height 300 -takefocus 0
    
    # First editor
    frame $w.1
    set e $w.1.txt
    grid [uplevel [list two_text $e -xs "$w.1.sx set" -ys "$w.1.sy set" -wrap none] $args] \
	 -row 0 -column 0 -sticky nsew
    grid [two_autosb $w.1.sy -command "$e yview"] \
	 -row 0 -column 1 -sticky ns
    grid [two_autosb $w.1.sx -command "$e xview" -orient horizontal] \
         -row 1 -column 0 -sticky ew
    grid rowconfigure    $w.1 0 -weight 1
    grid columnconfigure $w.1 0 -weight 1

    # really show scrollbar?
    if {!$Editor($w,scrollx)} {
	grid forget $w.1.sx
	$w.1.txt configure -wrap char
    }

    # bindings for matching of paranthesis
    if {$Editor($w,match)} {
	bindtags $e [concat [bindtags $e] TwoShowMatch]
    }
    $e tag configure match -background red -foreground white

    # Used for syntax highlighting
    $e tag configure string  -font twoEditorS -foreground \#00a
    $e tag configure comment -font twoEditorC -foreground \#080
    $e tag configure keyword -font twoEditorK -foreground \#800
    HighlightInit $w

    # Proxy events
    # Only do this with the first two_text widget to
    # avoid bindings being triggered multiple times
    bind $e <<ModOn>>   [list event generate $w <<ModOn>>]
    bind $e <<ModOff>>  [list event generate $w <<ModOff>>]
    bind $e <<SelOn>>   [list event generate $w <<SelOn>>]
    bind $e <<SelOff>>  [list event generate $w <<SelOff>>]
    bind $e <<UndoOn>>  [list event generate $w <<UndoOn>>]
    bind $e <<UndoOff>> [list event generate $w <<UndoOff>>]
    bind $e <<RedoOn>>  [list event generate $w <<RedoOn>>]
    bind $e <<RedoOff>> [list event generate $w <<RedoOff>>]
    bind $e <<PosChanged>> [list event generate $w <<PosChanged>> -x %x -y %y]

    # set active on focus in
    bind $w.1 <FocusIn> [namespace code "set Editor($w,active) $w.1.txt"]

    # pane
    two_splitter $w.1 -orient vertical -style $Editor($w,style)
    
    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Editor
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # initial state
    if {$Editor($w,split)} {
        Split $w
    }
    if {[string length $Editor($w,buffer)]} {
	New  $w $Editor($w,buffer)
    } elseif {[string length $Editor($w,file)]} {
        Open $w $Editor($w,file)
    } else {
        New $w
    }
    # Return the correct result...
    return $w
}

# ::two::editor::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w	Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::editor::Widget {w opt args} {
    variable Editor
    # See if this option is for us...
    switch -- $opt {
	configure {
	    set oldsplit $Editor($w,split)
	    two_parseArgs args Editor($w,closecmd)   "-closecmd"
	    two_parseArgs args Editor($w,saveerrcmd) "-saveerrcmd"
	    two_parseArgs args Editor($w,openerrcmd) "-openerrcmd"
	    two_parseArgs args Editor($w,split)      "-split"
	    two_parseArgs args Editor($w,comment)    "-comment"
	    two_parseArgs args Editor($w,string)     "-string"
	    two_parseArgs args Editor($w,keywords)   "-keywords"
	    two_parseArgs args Editor($w,keywordRE)  "-keywordRE"
	    HighlightInit $w
	    if {[two_parseArgs args Editor($w,highlight) "-highlight"]} {
	        if {!$Editor($w,highlight)} {
		    $w.1.txt tag remove string  1.0 end
		    $w.1.txt tag remove comment 1.0 end
		    $w.1.txt tag remove keyword 1.0 end
		} else {
		    HighlightStart $w
		}
	    }
	    if {[two_parseArgs args Editor($w,match) "-match"]} {
	        foreach ed $Editor($w,children) {
		    set b [bindtags $ed.txt]
		    if {[string match TwoShowMatch [lindex $b end]]} {
		        set b [lreplace $b end end]
		    }
        	    if {$Editor($w,match)} {
			set b [concat $b TwoShowMatch]
        	    }
		    bindtags $ed.txt $b
		}
	    }
	    if {[two_parseArgs args Editor($w,scrollx) "-scrollx"]} {
	        foreach ed $Editor($w,children) {
        	    if {$Editor($w,scrollx)} {
			grid $ed.sx -row 1 -sticky nsew
			$ed.txt configure -wrap none
        	    } else {
			grid forget $ed.sx
			$ed.txt configure -wrap char
		    }
		}
	    }
	    if {[two_parseArgs args Editor($w,file) "-file"]} {
	        Open $w $Editor($w,file)
	    }
	    if {[two_parseArgs args Editor($w,buffer) "-buffer"]} {
	        New $w $Editor($w,buffer)
	    }
	    if {[two_parseArgs args Editor($w,style) "-style"]} {
	        two_splitter configure $w.1 -style $Editor($w,style)
	    }
	    if {$Editor($w,split) && !$oldsplit} {
	        Split $w
	    } elseif {$oldsplit && !$Editor($w,split)} {
	        Join $w
	    }
	    if {[llength $args]} {
		return [eval [list $w.1.txt $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -closecmd   { return $Editor($w,closecmd)   }
	        -saveerrcmd { return $Editor($w,saveerrcmd) }
	        -openerrcmd { return $Editor($w,openerrcmd) }
	        -split      { return $Editor($w,split)      }
	        -scrollx    { return $Editor($w,scrollx)    }
	        -file       { return $Editor($w,file)       }
	        -buffer     { return $Editor($w,buffer)     }
	        -style      { return $Editor($w,style)      }
	        -match      { return $Editor($w,match)      }
	        -highlight  { return $Editor($w,highlight)  }
	        -comment    { return $Editor($w,comment)    }
	        -string     { return $Editor($w,string)     }
	        -keywords   { return $Editor($w,keywords)   }
	        -keywordRE  { return $Editor($w,keywordRE)  }
		-class      { return TwoEditor              }
		-takefocus  { return 0 }
		default     { return [uplevel $w.1.txt $opt $args] }
	    }
	}
	open {
	    Open $w [lindex $args 0]
	}
	close {
	    return [Close $w]
	}
	save {
	    Save $w [lindex $args 0]
	}
	fname -
	filename {
	    if {[string length $Editor($w,file)]} {
		return $Editor($w,file)
	    } else {
		return $Editor($w,buffer)
	    }
	}
	isNew {
            return $Editor($w,new)
	}
	isFile {
            if {[string length $Editor($w,file)]} {
                return 1
            } else {
                return 0
            }
	}
	isBuffer {
            if {[string length $Editor($w,buffer)]} {
                return 1
            } else {
                return 0
            }
	}
	setFocus {
	    if {[winfo exists $Editor($w,active)]} {
                focus $Editor($w,active)
	    } else {
                focus $w.1.txt
	    }
	}
	mtime {
	    return $Editor($w,mtime)
	}
	mtimeReset {
	    set file $Editor($w,file)
	    if {[string length $file]} {
                # We could use [file exists] and subsequently call
                # [file mtime], but this has a race condition (file could
                # have been removed in the meantime), so it's better to
                # use catch and just try to get the mtime.
		catch {set Editor($w,mtime) [file mtime $file]}
	    }
	}
	splitState {
	    return $Editor($w,split)
	}
	repaint {
	    if {$Editor($w,highlight)} {
	        HighlightStart $w
	    }
	}
	default {
	    # clones are almost equal, except for cursor
	    # position and scrolling. So we can't use
	    # $w.1.txt :-(
            return [uplevel [list $Editor($w,active) $opt] $args]
	}
    }
    return
}

# ::two::editor::_Destroy --
#
#	Call this procedure when the editor widget is destroyed.
#	Will clean widget-specific Elements of Editor and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::editor::_Destroy {w} {
    variable Editor

    # clear Editor
    foreach key [array names Editor "$w,*"] {
        unset Editor($key)
    }
    # remove the original widget command
    rename $w ""
}

#-------------------------------------------------------------------------
# Bindings
#-------------------------------------------------------------------------

# bindings for highlighting of matching paranthesis

# removed - doesn't seem to be useful
#bind TwoShowMatch <quotedbl>     {::two::editor::MatchBackward %W \" }
#bind TwoShowMatch <grave>        {::two::editor::MatchBackward %W \` }
#bind TwoShowMatch <apostrophe>   {::two::editor::MatchBackward %W \' }

bind TwoShowMatch <parenleft>      {::two::editor::MatchNestForw %W \( \) }
bind TwoShowMatch <parenright>     {::two::editor::MatchNestBack %W \) \( }

bind TwoShowMatch <braceleft>      {::two::editor::MatchNestForw %W \{ \} }
bind TwoShowMatch <braceright>     {::two::editor::MatchNestBack %W \} \{ }

bind TwoShowMatch <bracketleft>    {::two::editor::MatchNestForw %W \[ \] }
bind TwoShowMatch <bracketright>   {::two::editor::MatchNestBack %W \] \[ }

bind TwoShowMatch <Control-period> {::two::editor::ShowMatch %W}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::editor::New --
#
#	Create a new editor with no file; we call it a "buffer".
#
# Arguments:
#	w		Path to editor widget
#	name		Name of the new buffer

proc ::two::editor::New {w {name ""}} {
    variable Editor

    if {[string length $name]} {
        set Editor($w,new) 0
    } else {
        set name "New $Editor(nonce)"
        set Editor($w,new) 1
    }
    set  Editor($w,buffer) $name
    set  Editor($w,mtime)  0
    incr Editor(nonce)

    $w.1.txt afterLoad
}

# ::two::editor::Open --
#
#	Load file into buffer. If file == "" then a new filename is created.
#
# Argumets:
#	w		Path to editor widget
#	file		Name of file to open or empty for new file

proc ::two::editor::Open {w {file ""}} {
    variable Editor

    # first close old file -> give user a chance to abort
    if {[Close $w]} {
	if {[string length $file]} {
            set Editor($w,file) $file
            set Editor($w,new)  0
	    $w.1.txt delete 1.0 end

	    # the file commands might fail due to race conditions: we should
	    # catch such errors, but only if the widget user has provided an
	    # error handler. The widget is set to the new filename nevertheless
	    # to have a predictable state.
	    # {insert end} should never fail, so we can catch the whole block:-)
            if {[string length $Editor($w,openerrcmd)]} {
	        if {[catch {
                    set Editor($w,mtime) [file mtime $file]
	            $w.1.txt insert end [two_readFile $file]
                } errmsg]} {
                    uplevel "\#0" [list $Editor($w,openerrcmd) $file $errmsg]
                }
            } else {
                set Editor($w,mtime) [file mtime $file]
                $w.1.txt insert end [two_readFile $file]
            }

	    $w.1.txt mark set insert 1.0
            $w.1.txt afterLoad
	    $w.1.txt see 1.0

	    if {$Editor($w,highlight)} {
        	HighlightStart $w
	    }
	} else {
	    New $w
	}
    }
    return
}

# ::two::editor::Save --
#
#	Save buffer into file. If file == "" then the current name is used.
#
# Argumets:
#	w		Path to editor widget
#	file		Name of file to save as or empty for current name

proc ::two::editor::Save {w {file ""}} {
    variable Editor
    global errorCode

    if {[string length $file]} {
        set Editor($w,file) $file
        set Editor($w,new)  0
    } elseif {![string length $Editor($w,file)]} {
        return -code error "No filename set!"
    } else {
        set file $Editor($w,file)
    }

    # try twice if first attempt fails - work around temporary problems
    if {[catch {set fd [open $file "w"]} errmsg]} {
        after 200
        set err [catch {set fd [open $file "w"]} errmsg]
        if {$err} {
            # second try failed as well
            if {[string length $Editor($w,saveerrcmd)]} {
                uplevel "\#0" [list $Editor($w,saveerrcmd) $file $errmsg]
                return
            } else {
                return -code $err -errorcode $errorCode $errmsg
            }
        }
    }

    puts -nonewline $fd [$w.1.txt get 1.0 "end - 1 chars"]
    close $fd

    set Editor($w,mtime) [file mtime $file]
    $w.1.txt afterSave
    
    # Keyword Highlighting
    if {$Editor($w,highlight)} {
        HighlightStart $w
    }
    return
}

# ::two::editor::Close --
#
#	Delete contents of current buffer.
#	Returns 1 if file was really closes
#
# Argumets:
#	w		Path to editor widget

proc ::two::editor::Close {w} {
    variable Editor

    # run closecmd if there is one
    if {[string length $Editor($w,closecmd)]} {
        set retval [uplevel "\#0" [list $Editor($w,closecmd) $w]]
        if {$retval} {
	    # workaround: if {!"true"} breaks, but if {"true"} works
	} else {
	    return 0
	}
    }
    set Editor($w,file) ""
    $w.1.txt delete 1.0 end
    $w.1.txt afterLoad
    return 1
}

# ::two::editor::Split --
#
#	show second two_text widget and configure pane
#
# Argumets:
#	w		Path to widget

proc ::two::editor::Split {w} {
    variable Editor

    lappend Editor($w,children) $w.2

    frame $w.2
    set e $w.2.txt
    $w.1.txt clone $e
    $e configure -xs "$w.2.sx set" -ys "$w.2.sy set" -wrap none
    grid $e -row 0 -column 0 -sticky nsew
    grid [two_autosb $w.2.sy -command "$e yview"] \
	 -row 0 -column 1 -sticky ns
    grid [two_autosb $w.2.sx -command "$e xview" -orient horizontal] \
         -row 1 -column 0 -sticky ew
    grid rowconfigure    $w.2 0 -weight 1
    grid columnconfigure $w.2 0 -weight 1

    # bindings for matching of paranthesis
    if {$Editor($w,match)} {
	bindtags $e [concat [bindtags $e] TwoShowMatch]
    }

    # really show scrollbar?
    if {!$Editor($w,scrollx)} {
	grid forget $w.2.sx
	$w.2.txt configure -wrap char
    }

    # set active on focus in
    bind $w.2 <FocusIn> [namespace code "set Editor($w,active) $w.2.txt"]

    # add to pane and configure handle
    two_splitter $w.2

    # Signal
    event generate $w <<SplitOn>>
    return
}

# ::two::editor::Join --
#
#	show second two_text widget and configure pane
#
# Argumets:
#	w		Path to widget

proc ::two::editor::Join {w} {
    variable Editor

    set Editor($w,children) $w.1

    # has the second child the focus?
    if {[string match $w.2.* [focus]]} {
        # set focus to first child
        focus $w.1.txt
    }
    # is the second child marked as having had the focus recently?
    if {[string match $w.2.* $Editor($w,active)]} {
        set Editor($w,active) $w.1.txt
    }

    # remove from pane
    two_splitter forget $w.2
    # destroy second editor
    destroy $w.2
    # Signal
    event generate $w <<SplitOff>>
    return
}

# ::two::editor::MatchBackward --
#
#	Search for previous occurence of character "char" in
#	widget "w" and highlight both the character under
#	the insertion cursor and the matchingcharacter.
#	Useful for unnested matches.
#
# Status:
#	Deprecated
# 
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget
#	char -		Character which should match

proc ::two::editor::MatchBackward {w char} {
    variable Editor

    after cancel $Editor(matchId)
    $w tag remove match 1.0 end

    set found [$w search -backwards $char "insert -1c" 1.0]
    if {[string length $found]} {
	$w tag add match $found "$found +1c" "insert -1c" insert
	set Editor(matchId) [after 1000 "$w tag remove match 1.0 end"]
    }
    return
}

# ::two::editor::MatchNestForw --
#
#	Search for next matching parenthesis, bracket or
#	brace. Matching here also respects nesting of
#	opening and closing parenthesis etc. Useful for
#	binding to opening parenthesis.
#
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget
#	inc -		Character which increases the level of nesting
#	dec -		Character which decreases the level of nesting

proc ::two::editor::MatchNestForw {w inc dec} {
    variable Editor

    after cancel $Editor(matchId)
    $w tag remove match 1.0 end

    set idx insert
    set end [$w index "insert + 100 lines lineend"]
    set level 1
    set nextInc [$w search $inc $idx $end]
    set nextDec [$w search $dec $idx $end]
    while {1} {
        if {![string length $nextDec]} {
	    return
        }
	if {[string length $nextInc] &&
	    [$w compare "$nextInc +1c" < "$nextDec +1c"]
	} {
	    incr level
	    set nextInc [$w search $inc "$nextInc +1c" $end]
	} else {
	    incr level -1
	    if {!$level} break
	    set nextDec [$w search $dec "$nextDec +1c" $end]
	}
    }
    $w tag add match $nextDec "$nextDec +1c" "insert -1c" insert
    set Editor(matchId) [after 1000 "$w tag remove match 1.0 end"]
}

# ::two::editor::MatchNestBack --
#
#	Search for previous matching parenthesis, bracket or
#	brace. Matching here also respects nesting of
#	opening and closing parenthesis etc. Useful for
#	binding to closing parenthesis etc.
#
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget
#	inc -		Character which increases the level of nesting
#	dec -		Character which decreases the level of nesting

proc ::two::editor::MatchNestBack {w inc dec} {
    variable Editor

    after cancel $Editor(matchId)
    $w tag remove match 1.0 end

    set idx insert
    set end [$w index "insert - 100 lines linestart"]
    set level 1
    set nextInc [$w search -backwards $inc "$idx -1c" $end]
    set nextDec [$w search -backwards $dec "$idx -1c" $end]
    while {1} {
        if {![string length $nextDec]} {
	    return
        }
	if {[string length $nextInc] &&
	    [$w compare "$nextInc" > "$nextDec"]
	} {
	    incr level
	    set nextInc [$w search -backwards $inc "$nextInc" $end]
	} else {
	    incr level -1
	    if {!$level} break
	    set nextDec [$w search -backwards $dec "$nextDec" $end]
	}
    }
    $w tag add match $nextDec "$nextDec +1c" "insert -1c" insert
    set Editor(matchId) [after 1000 "$w tag remove match 1.0 end"]
}

# ::two::editor::ShowMatch --
#
#	Show matching parenthesis etc., depending character
#	at the insertion cursor (or the selected character).
#
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::ShowMatch {w} {
    if {[llength [$w tag ranges sel]]} {
        set char [$w get sel.first]
    } else {
        set char [$w get "insert -1c"]
    }
    switch $char {
        \" -
	\` -
	\' {
	    # removed - doesn't appear to be useful:-)
	    # MatchBackward $w $char
	}
	\( {
	    MatchNestForw $w \( \)
	}
	\) {
	    MatchNestBack $w \) \(
	}
	\[ {
	    MatchNestForw $w \[ \]
	}
	\] {
	    MatchNestBack $w \] \[
	}
	\{ {
	    MatchNestForw $w \{ \}
	}
	\} {
	    MatchNestBack $w \} \{
	}
    }
}

proc ::two::editor::lremove {listName element} {
    upvar $listName l
    
    set pos [lsearch -exact $l $element]
    if {$pos > -1} {
        set l [lreplace $l $pos $pos]
    }
}

# ::two::editor::HighlightInit --
#
#	Parse highlighting spec and build lists and hashtables to
#	increase the speed of the actual highlighting.
#
# Sideeffects:
#	Sets some entries to $Editor
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::HighlightInit {w} {
    variable Editor

    if {![llength $Editor($w,string)] && ![llength $Editor($w,comment)]} {
        return
    }

    # Comment and string markers are _no_ regular expression (they
    # are used as part of an array _key_ -> no regexp possible), so
    # we have to escape special characters for use within a regexp:
    set bad {([*+|()^\\])}

    # List of markers for start and stop of string or comment
    set startMarkerList ""
    foreach startMarker $Editor($w,string) {
	set Editor($w,tag,$startMarker) string
        set Editor($w,stop,$startMarker) $startMarker
        regsub -all -- $bad $startMarker {[\1]} startMarker
        lappend startMarkerList $startMarker
    }
    foreach {commentStart commentStop} $Editor($w,comment) {
        regsub -all -- $bad $commentStop {[\1]} commentStop
	set Editor($w,tag,$commentStart) comment
        set Editor($w,stop,$commentStart) $commentStop
        regsub -all -- $bad $commentStart {[\1]} commentStart
        lappend startMarkerList $commentStart
    }
    # build the regular expression
    set Editor($w,startRE) [join $startMarkerList |]
    if {[string length $Editor($w,keywords)]} {
        set Editor($w,fieldsep) 1
	append Editor($w,startRE) {|(^|[^_a-zA-Z0-9])(}
        append Editor($w,startRE) [join $Editor($w,keywords) |]
	append Editor($w,startRE) {)([^_a-zA-Z0-9]|$)}
    } else {
        set Editor($w,fieldsep) 0
    }
    if {[string length $Editor($w,keywordRE)]} {
        append Editor($w,startRE) |($Editor($w,keywordRE))
    }
    return
}

# ::two::editor::HighlightStart --
#
#	Parse current content and highlight according to current
#	config (comment, string and keywords). The real work is
#	done in HighlightLoop
#
# Sideeffects:
#	Adds tags to comments, strings and keywords in current widget
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::HighlightStart {w} {
    variable Editor

#if {![info exists Editor(hl,start)]} {
#    set Editor(hl,start) [clock clicks]
#}

    # First of all, stop previous highlighting within this
    # editor (if non is on the way, nothing will happen)
    HighlightStop $w
    # List of markers for start and stop of string or comment
    if {![llength $Editor($w,string)] && ![llength $Editor($w,comment)]} {
        return
    }
    event generate $w <<HighlightStart>>
    $w.1.txt tag remove string  1.0 end
    $w.1.txt tag remove comment 1.0 end
    $w.1.txt tag remove keyword 1.0 end
    # set cursor
    set Editor($w,cursor) [$w.1.txt cget -cursor]
    $w.1.txt configure -cursor watch
    
    # Add this widget to list of widgets for highlight loop
    lappend Editor(hl,widgets) $w
    set Editor($w,highlightPos) 1.0
    # schedule highlight loop anew if there is none right now
    if {!$Editor(hl,active)} {
        set Editor(hl,active) 1
        after idle "after 1 [namespace code HighlightLoop]"
    }
    return
}

# ::two::editor::HighlightStop --
#
#	Stop any undergoing highlighting for widget w
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::HighlightStop {w} {
    variable Editor

    if {$Editor(hl,active) && [lsearch -exact $Editor(hl,widgets) $w] > -1} {
        lremove Editor(hl,widgets) $w
	if {[winfo exists $w]} {
	    $w.1.txt configure -cursor $Editor($w,cursor)
            event generate $w <<HighlightStop>>
	}
    }
    return
}

# ::two::editor::HighlightLoop --
#
#	This is the loop for actually perfoming the highlighting.
#	It is done as an after-idle-loop to not freeze the user
#	interface while highlighting, but each iteration does
#	5 highlightings for efficiency reasons. Highlings are
#	done in all widgets given in the Editor(hl,widgets)
#	list.
#

proc ::two::editor::HighlightLoop {} {
    variable Editor

    # List of word characters
    set chars {[_a-zA-Z0-9]}
    # loop through all widgets which are to be highlighted
    foreach w $Editor(hl,widgets) {
        if {![winfo exists $w]} {
            # remove from list of highlighted windows
            lremove Editor(hl,widgets) $w
            continue
        }
    
        # cache RE in variable (avoid expensive Hashtable lookup)
        # and several other vars as well
        set startRE $Editor($w,startRE)
        set start $Editor($w,highlightPos)
        # We need to access the underlyign two_text widget. Usually,
        # this would be $w.txt.1 - but this goes through quite a lot
        # of code and makes the highlighting take about twice as long.
        # Non-modifying operations might well directly access the
        # TCL text widget, but modifications need to be done in all
        # clones (split editor!). Still it would be somehow better
        # to use $w.txt.1 as command, but we need the speed...
        # $txt1 -> one widget, only for non-modifying calls
        # $txt2 -> list of widgets; use whenever modifying the text
        set txt1 ::two::text::_two$w.1.txt
        if {$Editor($w,split)} {
            set txt2 [list ::two::text::_two$w.1.txt ::two::text::_two$w.2.txt]
        } else {
            set txt2 [list ::two::text::_two$w.1.txt]
        }
    
        # do 5 highlightings per widget at a time
        for {set loop 0} {$loop < 10} {incr loop} {
            # search for first candidate
            set start [$txt1 search -regexp \
                -count [namespace current]::Editor(hl,startLen) \
                -- $startRE $start end]
            if {![string length $start]} {
                # finished: there are no more candidates
                lremove Editor(hl,widgets) $w
                foreach txt $txt2 {
                    $txt configure -cursor $Editor($w,cursor)
                }
                event generate $w <<HighlightStop>>
                break
            }
        
            # what exactly is our candidate?
            set idx "$start + $Editor(hl,startLen) c"
            set match [$txt1 get $start $idx]
            if {[info exists Editor($w,stop,$match)]} {
                # O.k., we have a stop marker accompaining this start marker,
                # so it must be a string or a comment
                if {[IsEscaped $txt1 $start]} {
                    # escaped (backslashed) start delimiter -> ignore it:
                    # continue right after this delimiter
                    set start $idx
                } else {
                    # search for stop marker right after the start delimiter
                    while {1} {
                        # search for next candidate for stop delimiter.
                        # Need a workaround here: if stop delimiter is end-of-line,
                        # and $idx *is* at the end of the line, text search will
                        # return the end-of-line of the *next* line :-(
                        if {[string compare {$} $Editor($w,stop,$match)] == 0} {
                            set idx [$txt1 index "$idx lineend"]
                            set Editor(hl,stopLen) 0
                        } else {
                            set idx [$txt1 search -regexp \
                                -count [namespace current]::Editor(hl,stopLen) \
                                -- $Editor($w,stop,$match) $idx end]
                        }
                        if {[string length $idx]} {
                            if {$Editor(hl,stopLen)>0 && [IsEscaped $txt1 $idx]} {
                                # backslashed stop delimiter -> ignore it and
                                # continue right after the stop delimiter. We
                                # only accept this for non-zero length stop
                                # delimiters, so e.g. escaped end-of-lines are
                                # NOT covered here. There simply is no
                                # consistent to do this: different languages
                                # behave differently:-(
                                append idx " + $Editor(hl,stopLen) c"
                                continue
                            }
                            # found stop marker
                            set stop "$idx + $Editor(hl,stopLen) c"
                            foreach txt $txt2 {
                                $txt tag add $Editor($w,tag,$match) $start $stop
                            }
                            set start $stop
                            break
                        } else {
                            # there ain't no matching stop marker!?
                            # ignore this one and continue highlighting right after
                            # the start delimiter
                            append start " + $Editor(hl,startLen) c"
                            break
                        }
                    }
                }
            } else {
                # must be a keyword.
                if {$Editor($w,fieldsep)} {
                    # Keywords can be found at the beginning of the line or right after
                    # a field separator: in the second case, the field separator has
                    # been matched as well, but shouldn't get highlighted:-)
                    # Does the match start with a regular character (no field separator)?
                    # If not: set index l1 to 1
                    set l1 [expr {[string match $chars [string index $match 0]] ? 0 : 1}]
        
                    # Similar things happend at the end of a keyword: might be the lineend
                    # or a field separator.
                    # Does the match end with a regular character (no field separator)?
                    set len [string length $match]
                    incr len -1
                    set l2 [expr {[string match $chars [string index $match $len]] ? $len + 1 : $len}]
                } else {
                    # No list of keywords, but a RE given. Just ignore the fuzz
                    # about field separators: the RE should *exactly* describe
                    # the keyword (usually some kind of tag like in HTML)
                    set l1 0
                    set l2 [string length $match]
                }
                
                # add the tag and continue right after the keyword (excluding a
                # trailing field separator)
                set stop "$start + $l2 c"
                foreach txt $txt2 {
                    $txt tag add keyword "$start + $l1 c" $stop
                }
                set start $stop
            }
        }
        set Editor($w,highlightPos) $start
    }
    # give event loop a chance and then resort to the highlight loop
    if {[llength $Editor(hl,widgets)]} {
        after idle "after 1 [namespace code HighlightLoop]"
    } else {
        set Editor(hl,active) 0
#puts [expr {[clock clicks] - $Editor(hl,start)}]
#unset Editor(hl,start)
    }
    return
}

# ::two::editor::IsEscaped --
#
#	Test if the character at a given position is masked by a backslash.
#	It checks if there is an odd number of backslashes right before
#	this index (even would mean: backslash itself is escaped)
#
# Arguments:
#	txt		Path to text widget
#	idx		Position of possibly escaped character
#
# Returns:
#	1 if character is escaped, 0 otherwise.

proc ::two::editor::IsEscaped {txt idx} {
    if {[string compare \\ [$txt get "$idx -1c" $idx]]!=0} {
        return 0
    }
    set line [$txt get "$idx linestart" $idx]
    set pre  [string trimright $line \\]
    # We really should write "return [expr...]" here, but
    # this version should be faster:
    expr {([string length $line] - [string length $pre]) % 2}
}

namespace import ::two::editor::*
