# framework.tcl --
#
#	This file contains the basic procedure needed by the
#	TWO compound widgets framework.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/framework.tcl,v 1.5 1999/10/26 13:12:16 fjl Exp $

package provide twoFramework 1.0

namespace eval ::two::framework {
    namespace export two_*
    variable OptSpecs
    variable Aliases
    variable TemplateCreate
    variable TemplateManipulate

    # variable OptSpecs
    #	Array of Option descriptions for a widget. The key is "name,-opt"
    #	with name being the name of the widget (not the pathname!).
    #	The value is a list of 2 or 3 elements:
    #	{database} {class}
    #	{database} {class} {setProc}
    #	{database} {class} {setProc} {changeProc}
    
    array set OptSpecs ""
    
    # variable Aliases
    #	Array of Option aliases for a widget. The key is "name,-opt"
    #	with name being the name of the widget (not the pathname!).
    #	The value is the canonical name of the option.
    #	Additinal Entries are only "name" to list all aliases within
    #	the named widget
    
    array set Aliases ""
}

# variable ::two::framework::TemplateCreate --
#
#	This is the template for the procedure which will be created
#	by two_widget. It holds the setup code for creating new widgets.
#
# The following substitutions take place:
#	%state% -	Name of variable used to store widget config
#	%w% -		Name of widget that will be created (not the pathname!)
#	%base% -	Name of base widget
#	%init% -	Initinalize state array with default values
#	%create% -	Additional startup code
#	%options% -	List of all additinal options
#	%spec% -	Array of option specs (with fully qualified namespace)
#	%alias% -	Array of aliases (with fully qualified namespace)
#	%unknown% -	Handler proc for unknown options
#
# Pre-defined Variables:
#	Variables pre-defined for use in Template are $w (widget path) and $args.
#	There might be more from widget's option spec (fixed args).

set ::two::framework::TemplateCreate {
    variable %state%
    
    # Initialize the default values for widget config
    %init%

    # will contain all remaining options
    set remaining ""

    # keys will be the names all options that matched
    array set matched ""

    # loop through all options, but break at "--"
    set delimit [lsearch -exact $args "--"]
    if {$delimit > 0} {
        incr delimit -1
        set argList [lrange $args 0 $delimit]
    } else {
        set argList $args
    }
    foreach {arg value} $argList {
        if {[info exists %alias%(%w%,$arg)]} {
	    # just an alias to another option
	    set arg $%alias%(%w%,$arg)
	}
        if {[info exists %spec%(%w%,$arg)]} {
	    # real option: remember match
            set matched($arg) $value
	} else {
	    # left over for base widget
	    lappend remaining $arg $value
	}
    }
    # the rest....
    if {$delimit > 0} {
        lappend remaining [lrange $args $delimit end]
    }

    # build base widget
    %base% $w
    
    # set values in widget's state array
    # and query option database
    foreach arg %options% {
        set spec $%spec%(%w%,$arg)
        if {[info exists matched($arg)]} {
	    # this option was explicitly set
	    set %state%($w,$arg) $matched($arg)
	    if {[llength $spec] > 2} {
	        eval [lindex $spec 2]
	    }
	} else {
	    # query the option database
	    set opt [option get $w [lindex $spec 0] [lindex $spec 1]]
	    if {[string length $opt]} {
		set %state%($w,$arg) $opt
		if {[llength $spec] > 2} {
	            eval [lindex $spec 2]
		}
	    }
	}
    }

    # Rename the created base widget...
    rename $w _two$w

    # Configure the remaining options
    if [string length $remaining] {
	%unknown%
    }

    # ...and install our replacement command
    namespace eval :: "
        proc [list $w] {opt args} {
            uplevel [list %w%Manipulate] $w \$opt \$args
        }
    "

    # catch widget destruction in order to clean up Text
    bind $w <Destroy> [list [list %w%Clear] $w]
    
    # do additional startup processing
    %create%

    # Return the widget path
    return $w
}

# variable ::two::framework::TemplateManipulate --
#
#	This is the template for the procedure which will be created
#	used to manipulate the new widget created by two_text.
#
# The following substitutions take place:
#	%state% -	Name of variable used to store widget config
#	%w% -		Name by which the widget was created (not the pathname!)
#	%manip% -	switch body portion to match widget's commands
#	%options% -	List of all additinal options
#	%aliases% -	List of all additinal aliases
#	%spec% -	Array of option specs (with fully qualified namespace)
#	%alias% -	Array of aliases (with fully qualified namespace)
#	%aliases% -	List of all aliases for in the current widget's api
#	%default% -	Array of option's default values
#	%unknown% -	Handler proc for unknown options
#	%wns% -		Name of namespace in which the widget will be created
#
# Pre-defined Variables:
#	Variables pre-defined for use in Template are $w (widget path),
#	$cmd (widget command) and $args (additional arguments).

set ::two::framework::TemplateManipulate {
    variable %state%
    variable %default%

    # See if this option is for us...
    switch -exact -- $cmd {
        cget {
	    if {[llength $args] != 1} {
	        return -code error "Wrong # of args in \"$w cget $args\""
	    }
	    set arg [lindex $args 0]
            if {[info exists %alias%($w,$arg)]} {
		# just an alias to another option
		set arg $%alias%($w,$arg)
	    }
            if {[info exists %state%($w,$arg)]} {
		# known option
		return $%state%($w,$arg)
	    } else {
		# left over for base widget
		return [uplevel [list %wns%::_two$w] cget $arg]
	    }
	}
        conf -
	confi -
	config -
	configu -
	configur -
	configure {
	    if {![llength $args]} {
	        # return current configuration
	        set retval [_two$w configure]
		foreach opt %options% {
		    set spec $%spec%(%w%,$opt)
		    lappend retval [list \
		        $opt             \
			[lindex $spec 0] \
			[lindex $spec 1] \
			$%default%($opt)  \
			$%state%($w,$opt) \
		    ]
		}
		foreach opt $%alias%(%w%) {
		    set spec $%spec%(%w%,$%alias%(%w%,$opt))
		    lappend retval [list \
		        $opt             \
			[lindex $spec 0]
		    ]
		}
		return $retval
	    }
	    set remaining ""
	    foreach {arg value} $args {
        	if {[info exists %alias%(%w%,$arg)]} {
		    # just an alias to another option
		    set arg $%alias%(%w%,$arg)
		}
        	if {[info exists %spec%(%w%,$arg)]} {
		    set spec $%spec%(%w%,$arg)
		    # real option: react on it
		    set %state%($w,$arg) $value
		    if {[llength $spec] > 2} {
	        	eval [lindex $spec 2]
			if {[llength $spec] > 3} {
	        	    eval [lindex $spec 3]
			}
		    }
		} else {
		    # left over for base widget
		    lappend remaining $arg $value
		}
	    }
	    if {[string length $remaining]} {
	        return  [%unknown%]
	    }
	}
        %manip%
	default {
            return [uplevel [list %wns%::_two$w] $cmd $args]
	}
    }
    return
}

# variable ::two::framework::TemplateClear --
#
#	This is the template for the procedure which will
#	clean up state values on widget destruction.
#
# The following substitutions take place:
#	%state% -	Name of variable used to store widget config
#	%clear% -	Addional cleanup code
#
# Pre-defined Variables:
#	Variables pre-defined for use in Template is $w (widget path).

set ::two::framework::TemplateClear {
    variable %state%

    %clear%
    
    # clear state
    foreach key [array names %state% "$w,*"] {
        unset %state%($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

# ::two::framework::two_widget --
#
#	Utility procedure for the compound widget framework. Creates a
#	new procedure just like proc does, but with additinal option
#	parsing abilities (including Tk's Option Database), creating
#	and renaming of the "back-end" widget and setting preparing
#	for cleanup on widget destruction.
#	The optList is as list of option descriptions, each being a
#	a list themselves; following is a list of examples
#	{ {arg} }
#	{ {unknown} {proc} }
#	{ {-option} {cname} }
#	{ {-option} {database} {class} {default} }
#	{ {-option} {database} {class} {default} {setProc} }
#	{ {-option} {database} {class} {default} {setProc} {changeProc} }
#	The enties are:
#	arg -		A single fixed argument, like in normal proc def.
#	unknown -	Use this *keyword* to name the proc used
#			when changeing the configuration (.w configure)
#			and the option is not contained in this spec
#			This proc should be defined as "proc x {w args} {}"
#	-option -	The name of a option (including -)
#	cname -		Link to the canonical name of an option (NOT the class!).
#	database -	The name of the option in the option database
#	class -		The class of the option in the option database
#	default -	The default value
#	setProc -	Proc eval'ed whenever the value is set (or changed)
#	changeProc -	Proc eval'ed whenever the value is changed (configure)
#
# Arguments:
#	wname -		name of the compound widget procedure
#	stateVarName -	name of namespace variable carrying widget config
#	base -		name of base widget; additional options are appended!
#	optList -	List of lists describing the options (see manpage)
#	create -	code doing additional initialization
#	manipulate -	code used for the widget command
#	clear -		code used for additional clean-up
#
# Side Effects:
#	Creates a new procedure in caller's environment, the name
#	being the one given by "wname", and the argument list
#	"w args".
#
# Results:
#	Empty string.

proc ::two::framework::two_widget {
    wname stateVarName base optList
    {create     ""}
    {manipulate ""}
    {clear      ""}
} {
    variable OptSpecs
    variable Aliases
    variable DefaultsArray
    variable TemplateCreate
    variable TemplateManipulate
    variable TemplateClear
    
    set unknownOptHandler   {_two$w configure $remaining}
    array set DefaultsArray ""
    set Aliases($wname) ""
    set defaultsInit    ""
    set options         ""
    set fixedArgs       w
    
    # the namespace of the framework
    set fns [namespace current]
    # the namespace from which this proc has been called
    set cns [uplevel {namespace current}]
    # figure out in which namespace the proc should be created
    set wns [namespace qualifiers $wname]
    if {![string length $wns]} {
        if {![string match "::" $cns]} {
            set wns $cns
        }
    } elseif {![string match "::*" $wns]} {
        if {![string match "::" $cns]} {
            set wns $cns::$wns
        } else {
            set wns ::$wns
        }
    }

    # some sanity checks
    foreach {var content} [list wname $wname stateVarName $stateVarName base $base] {
	if {![string length $content]} {
            return -code error "value for \"$var\" missing"
	}
    }

    # transform the optList specification into a set of
    # lists and arrays in order to speed up later processing
    foreach spec $optList {
        switch [llength $spec] {
	    1 {
	      	# additional arg for widget command
		lappend fixedArgs [lindex $spec 0]
	    }
	    2 {
	    	# alias to another option or default proc for unknown option changes
	        if {[string match "unknown" [lindex $spec 0]]} {
		    set unknownOptHandler "[list eval [list [lindex $spec 1]]] \[list \$w\] \$remaining"
		} else {
		    if {![string match "-*" [lindex $spec 0]] } {
			return -code error "First element of two_widget optSpec \"$spec\" should start with \"-\""
		    }
		    if {![string match "-*" [lindex $spec 1]] } {
			return -code error "Alias (2. element) in two_widget optSpec \"$spec\" should start with \"-\""
		    }

		    lappend Aliases($wname) [lindex $spec 0]
		    set Aliases($wname,[lindex $spec 0]) [lindex $spec 1]
		}
	    }
	    4 -
	    5 -
	    6 {
	    	# option
	    	set opt [lindex $spec 0]

		if {![string match "-" [string index $opt 0]]} {
		    return -code error "First element of two_widget optSpec \"$spec\" should start with \"-\""
		}
		
    		# build array of option specs for later use
		# when parsing the actual options in the widget proc
		# The name is mangled into the key, and the default value removed
        	set OptSpecs($wname,$opt) \
		    [concat [lrange $spec 1 2] [lrange $spec 4 5]]

		# build list of options
		lappend options $opt
		
		# build list of default configs for array initialization
		append defaultsInit "\$w,$opt [list [lindex $spec 3]] \\\n        "
		set    DefaultsArray($opt)     [lindex $spec 3]
	    }
	    default {
		return -code error "Wrong # of args: should in two_widget optSpec \"$spec\""
	    }
	}
    }
    
    if {[string length $defaultsInit]} {
        set init "array set [list $stateVarName] \[list $defaultsInit\]"
    } else {
        set init "# none"
    }
    if {![string length $create]} {
        set create "# none"
    }

    # complete the creation procedure wproc
    set wproc ""
    foreach {code token} [split $TemplateCreate %] {
        append wproc $code
	switch -exact -- $token {
	    state   { append wproc $stateVarName      }
	    w       { append wproc $wname             }
	    base    { append wproc $base              }
	    init    { append wproc $init              }
	    options { append wproc [list $options]    }
	    spec    { append wproc ${fns}::OptSpecs   }
	    alias   { append wproc ${fns}::Aliases    }
	    unknown { append wproc $unknownOptHandler }
	    create  { append wproc $create            }
	}
    }

    # complete the manipulation procedure mproc
    set mproc ""
    foreach {code token} [split $TemplateManipulate %] {
        append mproc $code
	switch -exact -- $token {
            state   { append mproc $stateVarName         }
            w       { append mproc $wname                }
            options { append mproc [list $options]       }
            spec    { append mproc ${fns}::OptSpecs      }
            alias   { append mproc ${fns}::Aliases       }
            default { append mproc ${fns}::DefaultsArray }
            unknown { append mproc $unknownOptHandler    }
            manip   { append mproc $manipulate           }
	    wns     { append mproc $wns                  }
	}
    }

    # complete the cleanup procedure cproc
    set cproc ""
    foreach {code token} [split $TemplateClear %] {
        append cproc $code
	switch -exact -- $token {
	    state   { append cproc $stateVarName }
	    clear   { append cproc $clear        }
	}
    }

    # ...and finally create the proc
    # Note: $wname includes an absoulte namespace qualifier
    lappend fixedArgs args
    proc $wname             $fixedArgs   $wproc
    proc ${wname}Manipulate {w cmd args} $mproc
    proc ${wname}Clear      w            $cproc
}

namespace import ::two::framework::*
