# historybox.tcl --
#
#	Wrapper around the combobox widget to provide history
#	functionality: all commands entered will be remembered
#	and added to the drop down list. You can even navigate
#	the history with the cursor keys.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/historybox.tcl,v 1.4 1999/10/26 13:12:17 fjl Exp $

package provide twoHistorybox 1.0
package require twoUtil
package require twoCombobox

namespace eval ::two::historybox {
    namespace export two_historybox
    variable Historybox
    
    namespace import ::two::util::two_parseArgs
    
    # variable Historybox
    #
    # $w,current -	index of currently selected element from history
    # $w,max -	maximum number of entries in history
    # $w,hist -	list of previously entered commands
    # $w,command -	command to invoke when user presses "return"
    
    array set Historybox {
    }
}

# ::two::historybox::two_historybox --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.cb"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.cb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	combobox creates a toplevel widget "$w.drop" for the drop box.
#	Arguments are passed to entry widget.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::historybox::two_historybox {w args} {
    variable Historybox

    # Initialise the globals
    array set Historybox "
	$w,command {}
	$w,max     32
	$w,hist    {}
	$w,current -1
    "
    
    # See if this option is for us...
    two_parseArgs args Historybox($w,command) "-command"
    two_parseArgs args Historybox($w,max)     "-max"

    # the hard work is done by the combobox widget
    uplevel [list two_combobox $w] $args [list -command [namespace code [list SetList $w]]]

    # ... with the help of some bindings
    set e [$w entry]
    bind $e <Return> [list ::two::historybox::Invoke $w]
    bind $e <Up>     [list ::two::historybox::Prev   $w]
    bind $e <Down>   [list ::two::historybox::Next   $w]

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Historybox
    # BEWARE: Here we need to ADD our binding as the underlying
    # combobox has already added it's own destruction handler.
    bind $w <Destroy> +[namespace code [list _Destroy $w]]

    # Return the correct result...
    return $w
}

# ::two::historybox::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::historybox::Widget {w opt args} {
    variable Historybox

    # See if this option is for us...
    switch -exact -- $opt {
        add2hist {
            Add2hist $w
        }
        reset {
            set Historybox($w,current) -1
        }
	configure {
	    two_parseArgs args Historybox($w,command)  "-command"
	    if {[two_parseArgs args Historybox($w,max) "-max"]} {
	        Truncate $w
	    }

	    if {[llength $args]} {
		return [eval [list _two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -command { return $Historybox($w,command) }
	        -max     { return $Historybox($w,max)     }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::historybox::_Destroy --
#
#	Call this procedure when the Historybox widget is destroyed.
#	Will clean widget-specific Elements of Historybox.
#
# Arguments:
#	w		Path to Widget

proc ::two::historybox::_Destroy {w} {
    variable Historybox

    # clear Historybox
    foreach key [array names Historybox "$w,*"] {
        unset Historybox($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::historybox::Add2hist
#
#	The current contents gets appended to the history (if different from
#	the most recent entry)
#
# Arguments:
#	w		Path to widget

proc ::two::historybox::Add2hist {w} {
    variable Historybox

    set txt [_two$w get]

    # different from previous contents?
    if {[string compare $txt [lindex $Historybox($w,hist) 0]] != 0} {
        # prepend to history
        set Historybox($w,hist) [linsert $Historybox($w,hist) 0 $txt]
        Truncate $w
    }
    set History($w,current) -1
    return
}

# ::two::historybox::Invoke --
#
#	Evaluate the command ($w configure -command <cmd>). The current
#	contents also gets appended to the history (if different from
#	the most recent entry)
#
# Arguments:
#	w		Path to widget

proc ::two::historybox::Invoke {w} {
    variable Historybox

    Add2hist $w

    # any command to call?
    if {[string length $Historybox($w,command)]} {
        namespace eval :: $Historybox($w,command)
    }
    return
}

# ::two::historybox::Truncate --
#
#	Truncate the histrory to a maximum of ($w configure -max <max>)
#	elements.
#
# Arguments:
#	$w		Path to widget

proc ::two::historybox::Truncate {w} {
    variable Historybox
    
    set max $Historybox($w,max)
    incr max -1
    set Historybox($w,hist) [lrange $Historybox($w,hist) 0 $max]
    return
}

# ::two::historybox::SetList --
#
#	Set list for dropdown listbox from history
#
# Arguments:
#	w		Path to widget

proc ::two::historybox::SetList {w} {
    variable Historybox

    eval [list $w setlist] $Historybox($w,hist)
}

# ::two::historybox::Next --
#
#	go to next element of the history
#
# Arguments:
#	$w		Path to widget

# go to next element of the target history
proc ::two::historybox::Next {w} {
    variable Historybox

    if {$Historybox($w,current) > 0} {
	incr Historybox($w,current) -1
	_two$w set [lindex $Historybox($w,hist) $Historybox($w,current)]
    } else {
	set Historybox($w,current) -1
        _two$w set ""
    }
}

# ::two::historybox::Prev --
#
#	go to previous element of the history
#
# Arguments:
#	$w		Path to widget

proc ::two::historybox::Prev {w} {
    variable Historybox

    if {$Historybox($w,current) < [llength $Historybox($w,hist)] -1} {
	incr Historybox($w,current)
	_two$w set [lindex $Historybox($w,hist) $Historybox($w,current)]
    }
}

namespace import ::two::historybox::*
