# shell.tcl --
#
#	A very simple terminal emulation. The "shell" prompt, commands
#	which are to be run and their in-/output use different tags.
#	The command's name can be faked.
#
# Copyright (c) 1998-99 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/shell.tcl,v 1.34 1999/11/05 14:32:00 fjl Exp $

package provide twoShell 1.0
package require twoUtil

namespace eval ::two::shell {
    namespace export two_shell
    variable Shell
    
    namespace import ::two::util::two_parseArgs
    
    # variable Shell
    #
    # expect -	1 if expect is available, else 0
    # $w,prompt -	string to use as fake prompt
    # $w,spawnId -	used by expect
    # $w,tty -	name of terminal used by expect
    # $w,oldfocus -	remember focus when running a program
    # $w,aftercmd -	command to evaluate after completion
    # $w,destbs -   true if destructive backspace
    
    set Shell(expect) [expr {![catch {package require Expect}]}]
}

# ::two::shell::two_shell --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.sh"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.sh") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::shell::two_shell {w args} {
    variable Shell

    # Initialise the globals
    array set Shell "
	$w,prompt	{shell> }
	$w,spawnId	0
	$w,tty		{}
	$w,oldfocus	{}
	$w,destbs	1
    "

    # See if this option is for us...
    two_parseArgs args Shell($w,prompt) "-prompt"
    two_parseArgs args Shell($w,detsbs) "-destructivebs"

    # text widget
    uplevel [list text $w] $args -takefocus 0

    # Remove all bindings
    bindtags $w [list $w TwoShellTerm [winfo toplevel $w] all]

    # Insert prompt
    $w insert end $Shell($w,prompt) prompt
    $w see end
    
    # mark beginning of command
    $w mark set command end-1c
    $w mark gravity command left

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Shell
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::shell::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::shell::Widget {w opt args} {
    variable Shell

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    if {[two_parseArgs args Shell($w,prompt) "-prompt"]} {
	        # change all currently displayed prompts!
	        # The following works, because only one prompt per
	        # line is possible, so indexes are still valid after
	        # changing the first occurences.
	        foreach {start end} [_two$w tag ranges prompt] {
	            _two$w delete $start $end
	            _two$w insert $start $Shell($w,prompt) prompt
	        }
	    }
            two_parseArgs args Shell($w,destbs) "-destructivebs"
	    if {[llength $args]} {
		return [uplevel [list ::two::shell::_two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -prompt        { return $Shell($w,prompt) }
	        -destructivebs { return $Shell($w,destbs) }
		default        { return [uplevel [list ::two::shell::_two$w $opt] $args] }
	    }
	}
	run {
	    set Shell($w,aftercmd) ""
	    two_parseArgs args Shell($w,aftercmd) "-aftercmd"
	    # any command given?
	    if {![llength $args] || ![string length [lindex $args 0]]} {
	        return -code error "wrong # args: should be \"two_shell run command\""
	    }
	    # $fake is to be shown as command which is executed
	    set fake [lindex $args 0]
	    two_parseArgs args fake "-fake"
	    if {[llength $args] > 1} {
	        return -code error "unkown option \"[join [lrange $args 1 end] { }]\""
	    }
	    Run $w [lindex $args 0] $fake
	}
	kill {
	    Kill $w
	}
	runState {
	    if {$Shell($w,spawnId) != 0} {
        	return 1
	    } else {
        	return 0
	    }
	}
	clear {
	    $w delete 1.0 end
	    if {$Shell($w,spawnId) == 0} {
		$w insert end $Shell($w,prompt) prompt
	    }
	}
	default {
            return [uplevel [list ::two::shell::_two$w $opt] $args]
	}
    }
    return
}

# ::two::shell::_Destroy --
#
#	Call this procedure when the shell widget is destroyed.
#	Will clean widget-specific Elements of Shell and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::shell::_Destroy {w} {
    variable Shell

    # clear Shell
    foreach key [array names Shell "$w,*"] {
        unset Shell($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoShell-class bindings
#-------------------------------------------------------------------------

bind TwoShellActive <KeyPress> {
    ::two::shell::Input %W %A
}
bind TwoShellActive <Any-Tab> {
    ::two::shell::Input %W \t
    # don't change focus
    break
}

bind TwoShellActive <Control-c> {
    ::two::shell::Kill %W
}

bind TwoShellActive <Return> {
    if {$::two::shell::Shell(expect)} {
        ::two::shell::Input %W \r
    } else {
        ::two::shell::Input %W \n
    }
}

bind TwoShellActive <Button-1> {
    focus %W
}

# Special key sequences...
# they start with an escape: \x1b (decimal: 27)
bind TwoShellActive <Up>        {::two::shell::Input %W "\x1b\[A"}
bind TwoShellActive <Down>      {::two::shell::Input %W "\x1b\[B"}
bind TwoShellActive <Right>     {::two::shell::Input %W "\x1b\[C"}
bind TwoShellActive <Left>      {::two::shell::Input %W "\x1b\[D"}
bind TwoShellActive <End>       {::two::shell::Input %W "\x1b\[F"}
bind TwoShellActive <Home>      {::two::shell::Input %W "\x1b\[H"}
bind TwoShellActive <Insert>    {::two::shell::Input %W "\x1b\[2~"}
bind TwoShellActive <Delete>    {::two::shell::Input %W "\x1b\[3~"}
bind TwoShellActive <BackSpace> {::two::shell::Input %W "\b"}
bind TwoShellActive <Prior>     {::two::shell::Input %W "\x1b\[5~"}
bind TwoShellActive <Next>      {::two::shell::Input %W "\x1b\[6~"}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::shell::Input --
#
#	Insert text through currently running program
#
# Arguments:
#	w		Path to widget
#	txt		Text to send to program

proc ::two::shell::Input {w txt} {
    variable Shell

    if {[string length $txt]} {
        if {$Shell(expect)} {
            exp_send -i $Shell($w,spawnId) -- $txt
        } else {
            puts -nonewline $Shell($w,spawnId) $txt
            $w insert end $txt
            $w see end
        }
    }
}

# ::two::shell::Run --
#
#	execute a given command
#
# Argumets:
#	w		Path to widget
#	cmd		command to be executed
#	fake		display fake as command

proc ::two::shell::Run {w cmd fake} {
    variable Shell
    global env

    # remember focus
    set Shell($w,oldfocus) [focus]
    focus $w

    # already up and running?
    if {$Shell($w,spawnId) != 0} {
        Kill $w
    }

    # notify listener
    event generate $w <<RunningOn>>
    update idletasks

    # mark beginning of command
    $w mark set command end-1c

    # insert fake command
    $w insert end "$fake\n" command
    $w mark set insert end
    $w see insert
    
    if {$Shell(expect)} {
        # size of widget
        set f [$w cget -font]
        set rows [expr {[winfo reqheight $w] / [font metrics $f -linespace]}]
        set cols [expr {[winfo reqwidth $w]  / [font measure $f {0}]}]

	# clear env
	set envVars [list LINES COLUMNS TERM TERMCAP]
        foreach envVar $envVars {
            if {[info exists env($envVar)]} {
                set oldEnv($envVar) $env($envVar)
                unset env($envVar)
            }
        }
        set env(LINES)   $rows
        set env(COLUMNS) $cols
        # nd = Non-destructive space (cursor or carriage right)
        # cl = Clear screen and home cursor
        # ce = Clear to end of line

        # run command
        set stty_init "-tabs erase \^H columns $cols rows $rows"
        if {[catch [concat spawn -noecho -nottycopy $cmd]]} {
            $w insert end "Couldn't run "
            $w insert end "$fake\n" command
            Stopped $w
            return
        }
        set Shell($w,spawnId) $spawn_id
        set Shell($w,tty)     $spawn_out(slave,name)

        # don't echo anything on stdout
        log_user 0

	# re-set env to original state
        foreach envVar $envVars {
            if {[info exists oldEnv($envVar)]} {
                set env($envVar) $oldEnv($envVar)
            } elseif {[info exists env($envVar)]} {
                unset env($envVar)
            }
        }

        # read output from command
        expect_background "
            -i  $Shell($w,spawnId)
            eof {
                [namespace code Stopped] $w
            }
            {^\a} {
                bell
            }
            {^\b} {
                # Backspace
		if {$Shell($w,destbs)} {
		    $w delete {insert - 1c} insert
		} else {
                    $w mark set insert {insert - 1c}
		}
            }
            {^\n} {
                # Line feed
                $w mark set insert end
                $w insert insert {\n}
                $w see insert
            }
            {^\r} {
                # Carriage return
                $w mark set insert {insert linestart}
                $w see insert
            }
            -re {^(\[\x20-\xff\]|\t)+} {
                # char or tab (but tab shouldn't happen anyway)
                if {\[$w compare insert != {end -1c}\]} {
                    $w delete insert \"insert + \[string length \$expect_out(0,string)\] c\"
                }
                $w insert insert \$expect_out(0,string)
                $w see insert
            }
        "
    } else {
        # run command
        set Shell($w,spawnId) [open "|$cmd |& cat -u" r+]
        fconfigure $Shell($w,spawnId) -blocking 0 -buffering none

        # read output from command
        fileevent $Shell($w,spawnId) readable "
            if {\[eof $Shell($w,spawnId)\]} {
                [namespace code Kill] $w
            } else {
                $w insert end \[read $Shell($w,spawnId)\]
                $w see end
            }
        "
    }

    # Enable interaction
    bindtags $w [list $w TwoShellTerm TwoShellActive [winfo toplevel $w] all]
    _two$w configure -takefocus 1

    return
}

# ::two::shell::Kill --
#
#	stop execution of a command
#
# Argumets:
#	w		Path to widget

proc ::two::shell::Kill {w} {
    variable Shell

    if {$Shell($w,spawnId) != 0} {
        if {$Shell(expect)} {
            close -i $Shell($w,spawnId)
	} else {
            $w insert end [read $Shell($w,spawnId)]
	    catch {close $Shell($w,spawnId)}
	}
        Stopped $w
    }
    return
}

# ::two::shell::Stopped --
#
#	yes, command has been stopped. This procedure does all required
#	clean ups, state changes and evals the afterCommand
#
# Argumets:
#	w		Path to widget

proc ::two::shell::Stopped {w} {
    variable Shell
    global errorCode

    if {$Shell(expect) && $Shell($w,spawnId) != 0} {
        # Make sure the process is vanished -> avoid Zombies
        if {! [catch {set waitret [wait -i $Shell($w,spawnId)]}]} {
            # normal exciting with unusual return value?
            foreach {pid spid os ret errcl sig desc} $waitret {break}
#            if {$os == -1} {
#                $w insert end "OS error: $ret ($errorCode)\n" command
#            } elseif {$ret != 0} {
#                $w insert end "Return value: $ret\n" command
#            }
            if {[string match CHILDKILLED $errcl]} {
                $w insert end "Program killed: $desc ($sig)\n" command
            }
        }
    }
    set Shell($w,spawnId) 0

    # reset focus
    if {[string match $w [focus]] && [winfo exists $Shell($w,oldfocus)]} {
        focus $Shell($w,oldfocus)
    }

    # Make sure there is a newline
    if {![string match [$w index insert] [$w index "insert linestart"]]} {
        $w insert end "\n"
    }
    
    # Evaluate aftercmd (if set)
    if {[string length $Shell($w,aftercmd)]} {
        namespace eval :: $Shell($w,aftercmd)
    }

    # Show prompt
    $w insert end $Shell($w,prompt) prompt
    $w see end

    # Disable interaction
    bindtags $w [list $w TwoShellTerm [winfo toplevel $w] all]
    $w configure -takefocus 0
    if {[string match $w [focus]]} {
        focus [winfo parent $w]
    }

    event generate $w <<RunningOff>>
    return
}

namespace import ::two::shell::*
