# toolbar.tcl --
#
#	Provides a toolbar widget. Toolbars are usually placed below
#	an application's menue. They can hold several child widgets,
#	like all kinds of buttons, menus and comboboxes. They can
#	also be torn off.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/toolbar.tcl,v 1.41 1999/10/26 13:12:30 fjl Exp $

package provide twoToolbar 1.0
package require twoTooltip
package require twoUtil
package require twoCombobox

option add *TwoToolBar.Button.highlightThickness  0
option add *TwoToolBar.Button.takeFocus           0

# Workaround: Tk needs this procedure for cloning menus
# but won't load it form inside the namespace (bug in tcl8.0 up to p1)
catch {auto_load tkMenuDup}

namespace eval ::two::toolbar {
    namespace export two_toolbar
    variable Toolbar
    
    namespace import ::two::util::two_parseArgs
    
    # variable Toolbar
    #
    # nonce -	nonce to give any toplevel a new number
    # $w,nonce -	nonce to give any button within toolbar a new number
    # $w,tearoff -	If 1 then this toolbar can be torn off
    # $w,istorn -	1 if toolbar is currently torn of
    # $w,title -	Title string for torn-off toolbar
    # $w,#,state -	current state (on/off)
    # $w,#,value -	value for associated variable if button # is selected
    # $w,#,onvalue -	value for check'ed checkbutton
    # $w,#,offvalue -	value for uncheck'ed checkbutton
    # $w,#,variable -	variable associated with button
    # $w,#,command -	to be evalated on button press
    # $w,idx,"alias" -	index position of alias "alias"
    # $bt,state -	state of buttons
}

# Bitmap for tearoff handle
image create bitmap twoTearoff -foreground white -background black -data {
    \#define tearoff_main_width 8
    \#define tearoff_main_height 12
    static unsigned char tearoff_main_bits[] = {
      0x41, 0x08, 0x00, 0x41, 0x08, 0x00, 0x41, 0x08, 0x00, 0x41, 0x08, 0x00,
      };
} -maskdata {
    \#define tearoff_mask_width 8
    \#define tearoff_mask_height 12
    static unsigned char tearoff_mask_bits[] = {
      0x41, 0x8a, 0x10, 0x41, 0x8a, 0x10, 0x41, 0x8a, 0x10, 0x41, 0x8a, 0x10,
      };
}

# ::two::toolbar::two_toolbar --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.tb"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.tb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	To handle the tearoff, toolbar creates a second widget as toplevel
#	(path: "$w.top"). All options apply to both widgets.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::toolbar::two_toolbar {w args} {
    variable Toolbar

    # Initialise the globals
    array set Toolbar "
	$w,tearoff	1
	$w,istorn	0
	$w,nonce	0
	$w,row		$w.top.row0
	$w,title	Toolbar
    "

    # See if this option is for us...
    two_parseArgs args Toolbar($w,tearoff) "-tearoff"
    two_parseArgs args Toolbar($w,title)   "-title"

    # Container frame to be able to do a simple "pack forget" on whole toolbar
    # and restore frame again at the same position
    frame $w -borderwidth 0 -highlightthickness 0
    # Toolbar frame - this one takes the arguments
    uplevel [list frame $w.tb -class TwoToolBar] $args
    pack $w.tb -fill both -expand true
    # torn-off toolbar, takes the arguments as well
    uplevel [list toplevel $w.top -class TwoToolBar] $args
    wm withdraw  $w.top
    wm title     $w.top $Toolbar($w,title)
    wm transient $w.top $w.tb
    wm resizable $w.top 0 0
    wm protocol  $w.top WM_DELETE_WINDOW [namespace code "Tearoff $w 0"]
    wm withdraw  $w.top ;# needed on windows
    # Add tearoff buttons to both widgets
    button $w.tb.0  -image twoTearoff -command [namespace code "Tearoff $w 1"]
    button $w.top.0 -image twoTearoff -command [namespace code "Tearoff $w 0"]
    if {$Toolbar($w,tearoff)} {
	pack $w.tb.0  -side left -fill y
	pack $w.top.0 -side left -fill y
    }
    pack [frame $Toolbar($w,row) -bd 0] -side top -anchor w

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Toolbar
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::toolbar::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::toolbar::Widget {w opt args} {
    variable Toolbar

    if {[two_parseArgs args l "-label"]} {
        lappend args -text $l
    }

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    if {[two_parseArgs args Toolbar($w,title) "-title"]} {
	        wm title $w.top $Toolbar($w,title)
	    }
	    if {[two_parseArgs args Toolbar($w,tearoff) "-tearoff"]} {
                if {$Toolbar($w,tearoff)} {
                    pack $w.tb.0  -side left    -fill y
                    grid $w.top.0 -row 0 -col 0 -sticky ns
                } else {
                    pack forget $w.tb.0
                    grid forget $w.top.0
                }
	    }
	    if {[llength $args]} {
		return [eval [list _two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -tearoff { return $Toolbar($w,tearoff) }
	        -title   { return $Toolbar($w,title) }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	add {
	    if {[two_parseArgs args dummy "-off"]} {
		lappend args -state disabled
	    }
	    if {[two_parseArgs args dummy "-on"]} {
		lappend args -state normal
	    }
	    set type [lindex $args 0]
	    set args [lrange $args 1 end]
	    set tip  ""
	    set info ""
	    set doTip   [two_parseArgs args tip  "-tip"]
	    set doInfo  [two_parseArgs args info "-info"]
	    set index   [incr Toolbar($w,nonce)]
	    set buttons [list $w.tb.$index $w.top.$index]
	    if {[two_parseArgs args alias "-alias"]} {
		set Toolbar($w,idx,$alias) $index
	    }
	    switch -exact -- $type {
		separator -
		s {
	    	    foreach bt $buttons {
	            	uplevel [list frame $bt] $args -width 2 -relief sunken -bd 1
		    }
                    pack [lindex $buttons 0] -side left -fill y -padx 2 -pady 2
                    pack [lindex $buttons 1] -side left -fill y -padx 2 -pady 2 -in $Toolbar($w,row)
                }
                newline -
                n {
                    pack [uplevel [list frame [lindex $buttons 0]] $args -width 2 -relief sunken -bd 1] \
                        -side left -fill y -padx 2 -pady 2
                    pack [uplevel [list frame $w.top.sep$index] $args -height 2 -relief sunken -bd 1] \
                        -side top -fill x  -pady 2
                    set Toolbar($w,row) $w.top.row$index
                    pack [frame $Toolbar($w,row) -bd 0] -side top -anchor w
		}
		command -
		com -
		button -
		b {
	    	    foreach bt $buttons {
                        uplevel [list button $bt] $args -relief flat
                        bindtags $bt [list $bt TwoToolBarCommand Button [winfo toplevel $bt] all]
                    }
                    PackWidgets $w $buttons
		}
		combobox -
		combo {
	    	    foreach bt $buttons {
	            	uplevel [list two_combobox $bt] -takefocus 0 $args
		    }
                    PackWidgets $w $buttons
		}
		cascade -
		cas {
		    set slave ""
		    two_parseArgs args slave "-menu"
		    set state "normal"
		    if {[two_parseArgs args state "-state"]} {
			lappend args -state $state
		    }
	    	    foreach bt $buttons {
	            	uplevel [list menubutton $bt] $args [list -menu $bt.slave -relief flat]
		    	bindtags $bt [list $bt TwoToolBarMenubutton Menubutton [winfo toplevel $bt] all]
		        $slave clone $bt.slave
	        	set Toolbar($bt,state) $state
		    }
                    PackWidgets $w $buttons
		}
		checkbutton -
		check {
		    eval [list AddCheckbutton $w $index $buttons] $args
                    PackWidgets $w $buttons
		}
		radiobutton -
		radio {
		    eval [list AddRadiobutton $w $index $buttons] $args
                    PackWidgets $w $buttons
		}
		default {
		    incr Toolbar($w,nonce) -1
		    return -code error "unkown item type \"$type\""
		}
	    }
	    if {$doTip || $doInfo} {
	    	foreach bt $buttons {
	            two_tooltip $bt -tip $tip -info $info
		}
	    }
	    return $Toolbar($w,nonce)
	}
	entrybind {
	    set index    [lindex $args 0]
	    set sequence [lindex $args 1]
	    set script   [lindex $args 2]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    foreach bt "$w.tb.$index $w.top.$index" {
	        bind $bt $sequence [namespace code $script]
	    }
	}
        entrycget {
	    # TODO
	    set index  [lindex $args 0]
	    set option [lrange $args 1 end]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    if {$Toolbar($w,istorn)} {
		return [uplevel [list $w.top.$index cget] $option]
	    } else {
		return [uplevel [list $w.tb.$index cget] $option]
	    }
	}
	entrycommand -
	entrycmd {
	    set index  [lindex $args 0]
	    set option [lrange $args 1 end]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    if $Toolbar($w,istorn) {
		return [uplevel [list $w.top.$index] $option]
	    } else {
		return [uplevel [list $w.tb.$index] $option]
	    }
	}
        entryconfigure -
	entrycfg {
	    # TODO
	    if {[two_parseArgs args dummy "-off"]} {
		lappend args -state disabled
	    }
	    if {[two_parseArgs args dummy "-on"]} {
		lappend args -state normal
	    }
	    set state ""
	    if {[two_parseArgs args state "-state"]} {
		lappend args -state $state
	    }
	    set index   [lindex $args 0]
	    set options [lrange $args 1 end]
	    set tip     ""
	    set info    ""
	    set doTip   [two_parseArgs options tip  "-tip"]
	    set doInfo  [two_parseArgs options info "-info"]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    if {[two_parseArgs args alias "-alias"]} {
		set Toolbar($w,idx,$alias) $index
	    }
	    foreach bt "$w.tb.$index $w.top.$index" {
	        if {[string length $state]} {
	            set Toolbar($bt,state) $state
		}
		if {[llength $options]} {
		    uplevel [list $bt configure] $options
		}
		if {$doTip || $doInfo} {
	            two_tooltip $bt -tip $tip -info $info
		}
	    }
	}
	tearoff {
	    if {[llength $args] != 1} {
	        return -code error "wrong # args: should be \"two_toolbar tearoff boolean\""
	    }
            Tearoff $w [lindex $args 0]
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::toolbar::_Destroy --
#
#	Call this procedure when the toolbar widget is destroyed.
#	Will clean widget-specific Elements of Toolbar and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::toolbar::_Destroy {w} {
    variable Toolbar

    # destroy accompaining toplevel
    destroy $w.top
    # clear Toolbar
    set keys ""
    lappend keys [array names Toolbar "$w,*"]
    lappend keys [array names Toolbar "$w.tb.*,*"]
    lappend keys [array names Toolbar "$w.top.*,*"]
    foreach key $keys {
        unset Toolbar($key)
    }
    # remove the original widget command
    rename $w ""
    # remove trace procedures
    foreach p [info procs traceProc,$w,*] {
        rename $p ""
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoToolBar-class bindings
#-------------------------------------------------------------------------

# ---
# command button widgets
# ---
# raise button on mouse over

bind TwoToolBarCommand <Enter> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -relief raised
    }
}

bind TwoToolBarCommand <Leave> {
    %W configure -relief flat
}

# ---
# menu button widgets
# ---
# disable menubutton if not shown to enable proper
# handling of hotkeys (and F10)

bind TwoToolBarMenubutton <Map> {
    if {![string match "disabled" $::two::toolbar::Toolbar(%W,state)]} {
	%W configure -state normal
    }
}

bind TwoToolBarMenubutton <Unmap> {
    %W configure -state disabled
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::toolbar::Tearoff --
#
#	Tear off toolbar ("doit" == 1) or restore it ("doit" == 0)
#
# Arguments:
#	w		Path to Widget
#	doit		Tearoff (1) or restore (0)

proc ::two::toolbar::Tearoff {w doit} {
    variable Toolbar

    if {$doit} {
        set x [winfo rootx $w.tb]
        set y [winfo rooty $w.tb]
        wm geometry  $w.top +$x+$y
        update
        wm deiconify $w.top
	pack forget  $w.tb
	$w configure -height 1
	set Toolbar($w,istorn) 1
    } else {
	pack $w.tb -fill both -expand true
        wm withdraw $w.top
	set Toolbar($w,istorn) 0
    }
    # all elements are configured identically in both states
    # with combobox being the exception: it's too interactive:-)
    for {set index 0} {$index <= $Toolbar($w,nonce)} {incr index} {
        if {$Toolbar($w,istorn)} {
	    set b1 $w.tb.$index
    	    set b2 $w.top.$index
	} else {
	    set b1 $w.top.$index
    	    set b2 $w.tb.$index
	}
	set class ""
	catch {set class [$b1 cget -class]}
	if {[string match "TwoComboBox" $class]} {
	    eval [list $b2 setlist] [$b1 getlist]
	    eval [list $b2 set]     [$b1 get]
	}
    }
    return
}

# ::two::toolbar::PackWidgets --
#
#	Adds a widget to the toolbar, i.e. does the "pack" and "grid"
#	operations to add both versions of a widget to the toplevel or
#	the embedded version of the toolbar.
#
# Arguments:
#	w		Path to toolbar widget
#	buttons		widget names of the 2 widgets that should be placed
#			placed within the toolbar. First one is for the
#			embedded version, the second for the toplevel

proc ::two::toolbar::PackWidgets {w buttons} {
    variable Toolbar

    pack [lindex $buttons 0] -side left
    pack [lindex $buttons 1] -side left -in $Toolbar($w,row)
}

# ::two::toolbar::AddCheckbutton --
#
#	adds a checkbutton to the toolbar
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button
#	buttons		widget names of the 2 buttons that should be created
#	args		args of original widget command

proc ::two::toolbar::AddCheckbutton {w index buttons args} {
    variable Toolbar

    set Toolbar($w,$index,state)    0
    set Toolbar($w,$index,onvalue)  1
    set Toolbar($w,$index,offvalue) 0
    set Toolbar($w,$index,variable) ""
    set Toolbar($w,$index,command)  ""
    two_parseArgs args Toolbar($w,$index,onvalue)  "-onvalue"
    two_parseArgs args Toolbar($w,$index,offvalue) "-offvalue"
    two_parseArgs args Toolbar($w,$index,variable) "-variable"
    two_parseArgs args Toolbar($w,$index,command)  "-command"
    foreach bt $buttons {
	uplevel [list button $bt] $args [list -relief flat -command \
	    [list ::two::toolbar::CheckbuttonCommand $w $index]]
    }
    if {[string length $Toolbar($w,$index,variable)]} {
	upvar \#0 "$Toolbar($w,$index,variable)" var
	trace variable var w [namespace code traceProc,$w,$index]
	proc traceProc,$w,$index {name1 name2 op} "
	    variable Toolbar
            upvar \$name1 var
	    if {\[string length \$name2\]} {
		if {\[string match \$var(\$name2) \$Toolbar($w,$index,onvalue)\]} {
		    set Toolbar($w,$index,state) 1
		} elseif {\$var(\$name2) == \$Toolbar($w,$index,offvalue)} {
		    set Toolbar($w,$index,state) 0
		}
	    } else {
		if {\[string match \$var \$Toolbar($w,$index,onvalue)\]} {
		    set Toolbar($w,$index,state) 1
		} elseif {\[string match \$var \$Toolbar($w,$index,offvalue)\]} {
		    set Toolbar($w,$index,state) 0
		}
	    }
	    CheckbuttonState $w $index
	"
    }
    return
}

# ::two::toolbar::CheckbuttonCommand --
#
#	command called when checkbutton is pressed
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::CheckbuttonCommand {w index} {
    variable Toolbar
    set Toolbar($w,$index,state) [expr 1 - $Toolbar($w,$index,state)]
    if {[string length $Toolbar($w,$index,variable)]} {
        set var $Toolbar($w,$index,variable)
	uplevel \#0 "
	    if { $Toolbar($w,$index,state) } {
		if { ![info exists $var] || ![string match $$var $Toolbar($w,$index,onvalue)]} {
	            set $var $Toolbar($w,$index,onvalue)
		}
	    } else {
		if { ![info exists $var] || ![string match $$var $Toolbar($w,$index,offvalue)]} {
	            set $var $Toolbar($w,$index,offvalue)
		}
	    }
	"
    } else {
        # only call directly if there is no trace on var
        CheckbuttonState $w $index
    }
    if {[string length $Toolbar($w,$index,command)]} {
        uplevel \#0 $Toolbar($w,$index,command)
    }
    return
}

# ::two::toolbar::CheckbuttonState --
#
#	update state of checkbutton
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::CheckbuttonState {w index} {
    variable Toolbar
    foreach bt "$w.tb.$index $w.top.$index" {
	if {$Toolbar($w,$index,state)} {
            $bt configure -relief sunken
	} else {
            $bt configure -relief flat
	}
    }
    return
}

# ::two::toolbar::AddRadiobutton --
#
#	adds a radio button to the toolbar
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button
#	buttons		widget names of the 2 buttons that should be created
#	args		args of original widget command

proc ::two::toolbar::AddRadiobutton {w index buttons args} {
    variable Toolbar

    set Toolbar($w,$index,state)    0
    set Toolbar($w,$index,value)    $index
    set Toolbar($w,$index,variable) ""
    set Toolbar($w,$index,command)  ""
    two_parseArgs args Toolbar($w,$index,value)    "-value"
    two_parseArgs args Toolbar($w,$index,variable) "-variable"
    two_parseArgs args Toolbar($w,$index,command)  "-command"
    foreach bt $buttons {
	uplevel [list button $bt] $args [list -relief flat -command \
	    [list ::two::toolbar::RadiobuttonCommand $w $index]]
    }
    if {[string length $Toolbar($w,$index,variable)]} {
	set varname $Toolbar($w,$index,variable)
	upvar \#0 "$varname" var
	lappend Toolbar($w,radio,$varname) $index
	trace variable var w [namespace code traceProc,$w,$index]
	proc traceProc,$w,$index {name1 name2 op} "
	    variable Toolbar
            upvar \$name1 var
	    foreach i \$Toolbar($w,radio,$varname) {
		if {\[string length \$name2\]} {
		    if {\[string match \$var(\$name2) \$Toolbar($w,\$i,value)\]} {
			set Toolbar($w,\$i,state) 1
		    } else {
			set Toolbar($w,\$i,state) 0
		    }
		} else {
		    if {\[string match \$var \$Toolbar($w,\$i,value)\]} {
			set Toolbar($w,\$i,state) 1
		    } else {
			set Toolbar($w,\$i,state) 0
		    }
		}
		RadiobuttonState $w \$i
	    }
	"
    }
    return
}

# ::two::toolbar::RadiobuttonCommand --
#
#	command called when checkbutton is pressed. This will update
#	the global variable (if set) and set the state
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::RadiobuttonCommand {w index} {
    variable Toolbar
    set Toolbar($w,$index,state) [expr 1 - $Toolbar($w,$index,state)]
    if {[string length $Toolbar($w,$index,variable)]} {
        set var $Toolbar($w,$index,variable)
	uplevel \#0 "
	    if { ![info exists $var] || ![string match $$var $Toolbar($w,$index,value)]} {
	        set $var $Toolbar($w,$index,value)
	    }
	"
    } else {
        # only call directly if there is no trace on var
        RadiobuttonState $w $index
    }
    if {[string length $Toolbar($w,$index,command)]} {
        uplevel \#0 $Toolbar($w,$index,command)
    }
    return
}

# ::two::toolbar::RadiobuttonState --
#
#	update state of radiobutton according to contents of
#	the Toolbar variable
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::RadiobuttonState {w index} {
    variable Toolbar
    foreach bt "$w.tb.$index $w.top.$index" {
	if {$Toolbar($w,$index,state)} {
            $bt configure -relief sunken
	} else {
            $bt configure -relief flat
	}
    }
    return
}

namespace import ::two::toolbar::*
