# util.tcl --
#
#	This file contains several utility procedures used by
#	the two library. Only part of these are imported
#	into the global namespace.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/util.tcl,v 1.33 1999/11/08 20:26:24 fjl Exp $

package provide twoUtil 1.0

# Workaround: Tk needs this procedure for tk_messageBox
# but won't load it form inside a namespace (bug in tcl8.0 up to p1)
catch {auto_load tkMessageBox}

namespace eval ::two::util {
    namespace export two_*
    variable Util
    
    # variable Util
    #
    # clipboard -	Contents of replacement clipboard for two_noImportExport
    # PWD -		Current working directory for two_kshPwd
    # OLDPWD -	Previous working directory dor two_kshPwd
}

# ::two::util::two_parseArgs --
#
#	This procedures extracts an argument "arg" and a value
#	from "arg_list", removing both argument and value. The value
#	is stored in "val".
#
# Arguments:
#
#	argsListName	Call by Reference; name of tcl list with arguments
#	valName		Call by Reference; name of variable; takes the value
#	arg		The Argument to be extracted
#
# Results:
#	1 if "arg" was found in "arg_list", otherwise 0.

proc ::two::util::two_parseArgs {argsListName valName arg} {
    upvar $argsListName argsList
    upvar $valName val

    set pos [lsearch -exact $argsList $arg]
    set end [lsearch -exact $argsList "--"]
    if {$pos >= 0 && ($end < 0 || $pos < $end)} {
        set valpos [expr {$pos +1}]
        set val [lindex $argsList $valpos]
        set argsList [lreplace $argsList $pos $valpos]
	return 1
    }
    return 0
}

# ::two::util::two_parseMultiArgs --
#
#	This procedures examines a list of arguments (option-value pairs)
#	which were used when calling a procedure. All arguments whose
#	names are contained as keys in the specsArray will be removed. The
#	value of the specsArray is the name of a variable which should take
#	the value which is associated with the option.
#
# Arguments:
#
#	specsArrayName	Name of Array which should contain the specs for the
#			options
#	argsListName	Name of List of Arguments which should be processed
#
# Results:
#	List of all options which have been eaten by this proc.
#
# Side effects:
#	Modifies the contents of $argsListName.

proc ::two::util::two_parseMultiArgs {specsArrayName argsListName} {
    upvar $specsArrayName specsArray
    upvar $argsListName   argsList

    # will contain all options which have not been eaten by this proc
    set remaining ""
    # will contain the names all options that matched the description
    set matched ""

    # loop through all options, but break at "--"
    set length [llength $argsList]
    for {set idx 0} {$idx < $length} {incr idx} {
        set arg [lindex $argsList $idx]
	if {[string match "--" $arg]} {
	    eval lappend remaining [lrange $argsList $idx end]
	    break
	}
        if {[info exists specsArray($arg)]} {
	    lappend matched $arg
	    set next [lindex $argsList [incr idx]]
            uplevel [list set $specsArray($arg) $next]
	} else {
	    lappend remaining $arg
	}
    }
    set argsList $remaining
    return $matched
}

# ::two::util::two_getOpt --
#
#	Searches the database much the same way the "option get"
#	does. The return value is set to $default if the option
#	get command returns an empty string.
#
# Arguments:
#	w		Widget path
#	name		Name of option to te queried
#	class		Option class
#	default		Default value (if "option get" returns empty value)
#
# Results:
#	The option or the default value.

proc ::two::util::two_getOpt {w name class default} {
    set retval [option get $w $name $class]
    if {![string length $retval]} {
        set retval $default
    }
    return $retval
}

# ::two::util::two_noExportImport --
#
#	Disables all copying to and from the application. Disabling the
#	export function is easy through "exportSelection none", but
#	importing the seletion is a bit trickier. It is achieved here
#	through replacing the "selection" and "clipboard" commands.
#	Caveat: You'll have to call this function before instantiating
#	any widget in order to make the "option add" work.
#
# Side Effects:
#	Replaces the tcl clipboard and selection procedures with
#	fake ones. Those replacements intentinally do only work
#	within the application.

proc ::two::util::two_noExportImport {} {
    variable Util
   
    set Util(clipboard) ""

    # No exportSelection
    option add *exportSelection  false  100

    # Install our replacement commands
    proc ::selection {opt args} "
	return \[uplevel [namespace code FakeSelection] \$opt \$args \]
    "
    proc ::clipboard {opt args} "
	uplevel [namespace code FakeClipboard] \$opt \$args
    "
    return
}

# ::two::util::FakeSelection --
#
#	This is a replacement for the standard tk selection command.
#	See two_noExportImport

proc ::two::util::FakeSelection {opt args} {
    variable Util

    switch $opt {
        clear {
	    set Util(clipboard) ""
	}
        get {
	    return $Util(clipboard)
	}
    }
    return
}

# ::two::util::FakeClipboard --
#
#	This is a replacement for the standard tk clipboard command.
#	See two_noExportImport

proc ::two::util::FakeClipboard {opt args} {
    variable Util

    switch $opt {
        clear {
	    set Util(clipboard) ""
	}
        append {
	    while {[llength $args]} {
	        switch -glob -- [lindex $args 0] {
	            -- {
			append Util(clipboard) [join [lrange $args 1 end] ""]
			set args ""
		    }
		    -* {
			set args [lrange $args 2 end]
		    }
		    default {
			append Util(clipboard) [join $args ""]
			set args ""
		    }
		}
	    }
	}
	default {
	    return -code error "wrong # args: should be \"clipboard option ?arg arg ...?\""
	}
    }
    return
}

# ::two::util::two_kshPwd --
#
#	Workaround for pwd: don't show /tmp_mnt/....
#	This happens on systems using the automounter, which mounts file-
#	systems in /tmp_mnt/ and puts symlinks in place. The problem here
#	is that accessing a path through /tmp_mnt/ will not re-mount the
#	filesystem, and the tcl cd/pwd combination will always show this
#	/tmp_mnt/ path (e.g. in tk_getOpenFile).
#	This implementation behaves much like the Korn shell: whenever you
#	cd into a directory through symlinks it will remember the route
#	you took. Using "cd .." will go up one level in your route, i.e.
#	remove the last part of the pathname, and not go up to the "real"
#	parent directory. Please note that different UNIX shells behave
#	differently in this respect.
#
# Side effects:
#	The tcl cd and pwd commands are renamed into the current namespace
#	and wrapped with a replacement function.

proc ::two::util::two_kshPwd {} {
    variable Util
    global tcl_platform

    if {[info exists Util(PWD)]} {
        return -code error "two_kshPwd has already been installed!"
    }

    if {[string match "unix" $tcl_platform(platform)]} {
	set Util(PWD) [exec /bin/ksh -c "pwd"]
	set Util(OLDPWD) $Util(PWD)
	rename ::pwd _origPwd
	rename ::cd  _origCd
	proc ::pwd {} [namespace code {
	    return $Util(PWD)
	}]
	proc ::cd {args} "
	    return \[uplevel [namespace code KshCd] \$args\]
	"
    }
    return
}

# ::two::util::KshCd --
#
#	This is a replacement for the standard tk cd command.
#	See two_kshPwd

proc ::two::util::KshCd {{dir ""}} {
    variable Util
    global errorCode

    # special handling for "cd" and "cd -"
    if {![string length $dir]} {
        set dir ~
    } elseif {[string match "-" $dir]} {
        set dir $Util(OLDPWD)
    }
    set Util(OLDPWD) $Util(PWD) 

    if {[set _origCdRetCode [catch {_origCd $dir} _origCdRetVal]]} {
        return -code $_origCdRetCode -errorcode $errorCode $_origCdRetVal
    }
    if {[string match "absolute" [file pathtype $dir]]} {
        # Do tilde substitution
        set dir [file nativename $dir]
    }
    set lpwd [file split $Util(PWD)]
    foreach part [file split $dir] {
	switch -- $part {
	    . {
	    }
	    .. {
		set lpwd [lreplace $lpwd end end]
	    }
	    default {
		lappend lpwd $part
	    }
	}
    }
    if {[llength $lpwd]} {
        set Util(PWD) [eval file join $lpwd]
    } else {
        set Util(PWD) [file nativename /]
    }
    return $_origCdRetVal
}

# ::two::util::two_invoke --
#
#	Visually invoke button or set focus to next widget if this function
#	is used on a label.
#
# Arguments:
#	w		Widget path to button
#
# Side effects:
#	The button is flashed and the invoked, or the focus is set to the
#	next widget.

proc ::two::util::two_invoke {w} {
    switch -exact -- [winfo class $w] {
	Button {
	    if {![string match "disabled" [$w cget -state]]} {
    		$w configure -state active -relief sunken
    		update idletasks
    		after 100
    		$w configure -state normal -relief raised
		$w invoke
	    }
	}
	Label {
	    set next [tk_focusNext $w]
	    if {[string match Entry [winfo class $next]]} {
		$next selection range 0 end
		$next icursor end
	    }
	    focus $next
	}
	default {
            $w invoke
	}
    }
    return
}

# ::two::util::two_push --
#
#	Push value on stack. Stack is a tcl list. Returns value.
#
# Arguments:
#	stackName	Name of TCL list variable
#	value		Value to be pushed on the stack
#
# Results:
#	The element which is pushed on the stack is returned as well.


proc ::two::util::two_push {stackName value} {
    upvar 1 $stackName stack

    lappend stack $value
    return $value
}

# ::two::util::two_pop --
#
#	Take topmost value from on stack. Stack is a tcl list. Returns value.
#
# Arguments:
#	stackName	Name of TCL list variable
#
# Results:
#	The topmost element is returned.

proc ::two::util::two_pop {stackName} {
    upvar 1 $stackName stack
    
    if {![info exists stack]} {
        return {}
    }
    set value [lindex $stack end]
    set stack [lreplace $stack end end]
    return $value
}

# ::two::util::two_now --
#
#	returns current time as formatted string
#
# Results:
#	the current time...

proc ::two::util::two_now {} {
    return [clock format [clock seconds]]
}

# ::two::util::two_ct --
#
#	returns the current toplevel, i.e. the toplevel which
#	currently contains the focus. If the application doesn't
#	have the focus, "." is returned
#
# Results:
#	A widget path

proc ::two::util::two_ct {} {
    set f [focus]
    if {[string length $f]} {
        return [winfo toplevel $f]
    } else {
        return "."
    }
}

# ::two::util::two_mb --
#
# thin wrapper around tk_messageBox which sets the parent widget to [two_ct]
#
# Arguments:
#	See man page

proc ::two::util::two_mb {args} {
    if {![two_parseArgs args p "-parent"] && ![two_parseArgs args p "-p"]} {
        set p [two_ct]
    }
    return [uplevel tk_messageBox -parent $p $args]
}

# ::two::util::two_center --
#
#	center widget w on top of parent p

proc ::two::util::two_center {w {p ""}} {
    wm withdraw  $w
    if {[string length $p]} {
        set p [winfo toplevel $p]
        wm deiconify $p
    } else {
        set p .
    }
    # center on screen
    update idletasks
    set rw [winfo reqwidth  $w]
    if {[winfo width $w] > $rw} {
        set rw [winfo width $w]
    }
    set rh [winfo reqheight $w]
    if {[winfo height $w] > $rh} {
        set rh [winfo height $w]
    }
    set x [expr {[winfo screenwidth  $p]/2 - $rw/2 - [winfo vrootx $p]}]
    set y [expr {[winfo screenheight $p]/2 - $rh/2 - [winfo vrooty $p]}]
    wm geom      $w ${rw}x${rh}+$x+$y
    wm deiconify $w
    return
}

# ::two::util::two_show --
#
#	show window w (i.e. raise/deiconify/show centered)

proc ::two::util::two_showCenter {w {p ""}} {

    if {[winfo ismapped $w]} {
        raise $w
    } elseif {[string match "iconic" [wm state $w]]} {
        wm deiconify $w
    } else {
        two_center $w $p
    }
    return
}

# ::two::util::two_busy
#
#	block interaction with a toplevel and show a "watch" cursor.
#	This function uses a transparent frame which is placed on
#	top of the toplevel widget.
#
# Know Bug:
#	Transparent frames don't propagate redraw events, so the
#	underlying widgets don't repaint at all. Looks like a
#	blocked application. Won't make any difference if your
#	app really blocks, but if it doesn't... it's ugly.
#
# Arguments:
#	busy		True to make the toplevels busy, false to
#			remove the busy cursor.
#	args		List of Paths to widgets which should be
#			disabled (preferrably toplevels)

proc ::two::util::two_busy {busy args} {
    foreach w $args {
        if {[string compare "." $w] == 0} {
            set f .twoBusyCursorFrame
        } else {
            set f $w.twoBusyCursorFrame
        }
        if {$busy} {
            if {[winfo exists $f]} {
                raise $f
            } else {
                frame $f -background {} -borderwidth 0 -cursor watch
                place $f -x 0 -y 0 -relwidth 1 -relheight 1
            }
        } else {
            if {[winfo exists $f]} {
                destroy $f
            }
        }
    }
    update idletasks
}


# ::two::util::two_readFile --
#
#	Read a given a file and return it's contents. This proc
#	combines the open, read and close calls into one short
#	procedure; errors are "forwarded" (use catch...). One
#	other thing that is handled within this proc is the
#	possibility of slow automounters etc. making an open
#	command fail. In any case, the open is tried twice
#	withing 200 ms.
#
# Arguments:
#	args		?-nonewline? ?-binary? fname
#
# Returns:
#	The contents of this file

proc ::two::util::two_readFile {args} {
    global errorCode

    # parse options
    set fname [lindex $args end]
    set nonewline 0
    set bin 0
    set opts [lreplace $args end end]
    foreach opt $opts {
        switch -exact -- $opt {
            -nonewline {
                set nonewline 1
            }
            -binary {
                set bin 1
            }
            default {
                return -code error "unknown option \"$opt\": must be -nonewline or -binary"
            }
        }
    }

    # Open file: try twice
    set err [catch {set fid [open $fname]}]
    if {$err} {
        # didn't work... try again after 200 ms to work around slow automounter
	after 200
        set err [catch {set fid [open $fname]} errorMsg]
        if {$err} {
            # second try failed as well
            return -code $err -errorcode $errorCode $errorMsg
        }
    }

    # read the file
    # Improve performance: TCL's default buffer size is 4096 and this is used
    # to read the file in "slices" and append them to the txt variable. There
    # are two methods to optimize this: set buffersize to something larger
    # (but it needs to be smaller than 1,000,000), e.g. 2^18, or use the real
    # size of the file as parameter to the open command (only available if
    # -nonewline ain't specified). Timing on a HP C160 (HP-UX 10.20) with two
    # large files (/stand/vmunix: ~6M, /opt/aCC/lib/lib++.a ~1M) reveal:
    #                   6M       1M
    # plain:      64058498  1744060
    # fconfigure:  2661062   151780
    # file size:    794698   115280
    if {$bin} {
        fconfigure $fid -translation binary
    }
    if {$nonewline} {
        fconfigure $fid -buffersize 262144
        set err [catch {set txt [read -nonewline $fid]} errorMsg]
    } else {
        set err [catch {set fs [file size $fname]} errorMsg]
        if {! $err} {
            set err [catch {set txt [read $fid $fs]} errorMsg]
        }
    }
    
    # closing should always work
    close $fid
    # handle error from above (reading file)
    if {$err} {
        return -code $err -errorcode $errorCode $errorMsg
    }
    
    # usually you would insert "return $txt" at this palce, but the following
    # should be faster (said to be due to less error checking?)
    set txt
}

namespace import ::two::util::two_noExportImport
namespace import ::two::util::two_kshPwd
namespace import ::two::util::two_invoke
namespace import ::two::util::two_ct
namespace import ::two::util::two_mb
namespace import ::two::util::two_center
namespace import ::two::util::two_showCenter
namespace import ::two::util::two_busy
namespace import ::two::util::two_readFile
